#!/usr/bin/env python

import os
from unittest import TestCase, main

from g2p import make_g2p
from g2p.log import LOGGER
from g2p.mappings import Mapping
from g2p.tests.public import __file__ as public_data
from g2p.transducer import Transducer


class LexiconTransducerTest(TestCase):
    def test_lexicon_mapping(self):
        """Test loading a lexicon mapping directly in the constructor."""
        with self.assertLogs(LOGGER, level="INFO"):
            m = Mapping(
                type="lexicon",
                case_sensitive=False,
                out_delimiter=" ",
                alignments_path=os.path.join(
                    os.path.dirname(public_data), "mappings", "hello.aligned.txt"
                ),
            )
        self.assertEqual(m.rules, [])
        self.assertEqual(m.type, "lexicon")
        t = Transducer(m)
        tg = t("hello")
        self.assertEqual(tg.output_string, "HH EH L OW ")
        self.assertEqual(
            tg.edges, [(0, 0), (0, 1), (1, 3), (1, 4), (2, 6), (3, 6), (4, 8), (4, 9)]
        )
        tg = t("you're")
        self.assertEqual(tg.output_string, "Y UH R ")
        self.assertEqual(
            tg.edges,
            [(0, 0), (1, 2), (1, 3), (2, 2), (2, 3), (3, 4), (4, 5), (5, 5)],
        )
        # These alignments are somewhat bogus, hence the name
        tg = t("bogus")
        self.assertEqual(tg.input_string, "bogus")
        self.assertEqual(tg.output_string, "")
        self.assertEqual(
            tg.edges,
            [(0, None), (1, None), (2, None), (3, None), (4, None)],
        )
        self.assertEqual(
            tg.substring_alignments(),
            [("bogus", "")],
        )
        tg = t("bogus")
        tg += t("hello")
        self.assertEqual(tg.input_string, "bogushello")
        self.assertEqual(tg.output_string, "HH EH L OW ")
        self.assertEqual(
            tg.edges,
            [
                (0, 0),
                (1, 0),
                (2, 0),
                (3, 0),
                (4, 0),
                (5, 0),
                (5, 1),
                (6, 3),
                (6, 4),
                (7, 6),
                (8, 6),
                (9, 8),
                (9, 9),
            ],
        )
        self.assertEqual(
            tg.substring_alignments(),
            [("bogush", "HH"), ("e", "EH"), ("ll", "L"), ("o", "OW")],
        )
        tg = t("hello")
        tg += t("bogus")
        tg += t("you're")
        tg += t("bogus")
        self.assertEqual(tg.input_string, "hellobogusyou'rebogus")
        self.assertEqual(
            tg.edges,
            [
                (0, 0),
                (0, 1),
                (1, 3),
                (1, 4),
                (2, 6),
                (3, 6),
                (4, 8),
                (4, 9),
                (5, 9),
                (6, 9),
                (7, 9),
                (8, 9),
                (9, 9),
                (10, 11),
                (11, 13),
                (11, 14),
                (12, 13),
                (12, 14),
                (13, 15),
                (14, 16),
                (15, 16),
                (16, 16),
                (17, 16),
                (18, 16),
                (19, 16),
                (20, 16),
            ],
        )
        self.assertEqual(
            tg.substring_alignments(),
            [
                ("h", "HH"),
                ("e", "EH"),
                ("ll", "L"),
                ("obogus", "OW"),
                ("y", "Y"),
                ("ou", "UH"),
                ("'", " "),
                ("rebogus", "R"),
            ],
        )

    def test_load_lexicon_mapping(self):
        """Test loading a lexicon mapping through a config file."""
        with self.assertLogs(LOGGER, level="INFO"):
            m = Mapping.load_mapping_from_path(
                os.path.join(
                    os.path.dirname(public_data), "mappings", "lexicon_config-g2p.yaml"
                )
            )
        self.assertEqual(m.rules, [])
        self.assertEqual(m.type, "lexicon")
        t = Transducer(m)
        tg = t("hello")
        self.assertEqual(tg.output_string, "HH EH L OW ")
        self.assertEqual(
            tg.edges, [(0, 0), (0, 1), (1, 3), (1, 4), (2, 6), (3, 6), (4, 8), (4, 9)]
        )

    def test_bad_lexicon_mapping(self):
        """Test failure to load alignments."""
        with self.assertRaises(FileNotFoundError), self.assertLogs(
            LOGGER, level="INFO"
        ):
            _ = Mapping.load_mapping_from_path(
                os.path.join(
                    os.path.dirname(public_data),
                    "mappings",
                    "bad_lexicon_config-g2p.yaml",
                )
            )

    def test_eng_lexicon(self):
        """Test the cached eng to eng-ipa lexicon as a Mapping."""
        m = Mapping.find_mapping(in_lang="eng", out_lang="eng-ipa")
        self.assertEqual(m.type, "lexicon")
        t = Transducer(m)
        tg = t("hello")
        self.assertEqual(tg.output_string, "hʌloʊ")
        self.assertEqual(tg.edges, [(0, 0), (1, 1), (2, 2), (3, 2), (4, 3), (4, 4)])
        tg = t("you're")
        self.assertEqual(tg.output_string, "jʊɹ")
        self.assertEqual(tg.edges, [(0, 0), (1, 0), (2, 1), (3, 1), (4, 2), (5, 2)])
        tg = t("change")
        self.assertEqual(tg.output_string, "tʃeɪndʒ")
        self.assertEqual(tg.input_string, "change")
        self.assertEqual(
            tg.edges,
            [
                (0, 0),
                (0, 1),
                (1, 1),
                (2, 2),
                (2, 3),
                (3, 4),
                (4, 5),
                (4, 6),
                (5, 6),
            ],
        )
        tg = t("chain")
        # These aligments are weird but they are the ones EM gave us
        # (and the ones that our arbitrary assignment of deletions to
        # adjacent outputs gives us, too...)
        self.assertEqual(tg.output_string, "tʃeɪn")
        self.assertEqual(tg.input_string, "chain")
        self.assertEqual(tg.edges, [(0, 0), (0, 1), (1, 1), (2, 2), (3, 3), (4, 4)])
        tg = t("xtra")
        self.assertEqual(tg.output_string, "ɛkstɹʌ")
        self.assertEqual(tg.input_string, "xtra")
        self.assertEqual(tg.edges, [(0, 0), (0, 1), (0, 2), (1, 3), (2, 4), (3, 5)])
        pe = tg.pretty_edges()
        self.assertEqual(
            pe,
            [("x", "ɛ"), ("x", "k"), ("x", "s"), ("t", "t"), ("r", "ɹ"), ("a", "ʌ")],
        )

    def test_eng_transducer(self):
        """Test the cached eng to eng-ipa lexicon from make_g2p."""
        transducer = make_g2p("eng", "eng-arpabet")
        tg = transducer("hello")
        self.assertEqual(tg.output_string, "HH AH L OW ")

        # since we tokenize by default now, this works:
        self.assertEqual(
            transducer("hello my friend").output_string, "HH AH L OW  M AY  F R EH N D "
        )

    def test_eng_lexicon_corner_cases(self):
        """White-box testing for compact storage of lexicon mappings."""
        test_cases = (
            ("'bout", "baʊt"),  # first entry in eng->eng-ipa
            ("'cause", "kʌz"),  # second entry
            ("'course", "kɔɹs"),  # third
            ("'tis", "tɪz"),  # 15th entry
            ("'twas", "twʌz"),  # 16th entry
            ("a", "ʌ"),  # 17th entry
            ("buttering", "bʌtɜ˞ɪŋ"),  # 15998th, which is -2 mod 16
            ("buttermilk", "bʌtɜ˞mɪlk"),  # 15999th, -1 mod 16
            ("buttermore", "bʌtɜ˞mɔɹ"),  # 16000th, 0 mod 16
            ("butters", "bʌtɜ˞z"),  # 16001th, 1 mod 16
            ("butterscotch", "bʌtɜ˞skɑtʃ"),
            ("butterworth", "bʌtɜ˞wɜ˞θ"),
            ("buttery", "bʌtɜ˞i"),
            ("butthead", "bʌthɛd"),
            ("butting", "bʌtɪŋ"),
            ("buttitta", "butitʌ"),
            ("buttke", "bʌtki"),
            ("buttler", "bʌtlɜ˞"),
            ("buttner", "bʌtnɜ˞"),
            ("buttock", "bʌtʌk"),
            ("buttocks", "bʌtʌks"),
            ("button", "bʌtʌn"),
            ("buttoned", "bʌtʌnd"),
            ("buttonhole", "bʌtʌnhoʊl"),
            ("buttonholed", "bʌtʌnhoʊld"),
            ("buttonholes", "bʌtʌnhoʊlz"),
            ("buttons", "bʌtʌnz"),  # 16018th
            ("zwieg", "zwiɡ"),  # last block of the lexicon
            ("zwilling", "zwɪlɪŋ"),
            ("zwolinski", "zvʌlɪnski"),
            ("zycad", "zɪkæd"),
            ("zych", "zaɪtʃ"),
            ("zycher", "zɪkɜ˞"),
            ("zydeco", "zaɪdʌkoʊ"),
            ("zygmunt", "zɪɡmʌnt"),
            ("zygote", "zaɪɡoʊt"),
            ("zyla", "zɪlʌ"),
            ("zylka", "zɪlkʌ"),
            ("zylstra", "zɪlstɹʌ"),
            ("zyman", "zaɪmʌn"),
            ("zynda", "zɪndʌ"),
            ("zysk", "zaɪsk"),
            ("zyskowski", "zɪskɔfski"),
            ("zyuganov", "zjuɡɑnɑv"),
            ("zyuganov's", "zjuɡɑnɑvz"),
            ("zywicki", "zɪwɪki"),
        )

        transducer = make_g2p("eng", "eng-ipa", tokenize=False)
        for word, expected in test_cases:
            tg = transducer(word)
            self.assertEqual(tg.output_string, expected)
            before = word[:-1] + chr(ord(word[-1]) - 1) + "z"
            self.assertEqual(
                transducer(before).output_string, "", f"word={word} before={before}"
            )
            after = word[:-1] + chr(ord(word[-1]) + 1) + "z"
            self.assertEqual(
                transducer(after).output_string, "", f"word={word} after={after}"
            )


if __name__ == "__main__":
    main()



"""

Tokenizer module, to provide language-dependent tokenization of text.
A token is defined as a sequence of characters that are either part of the
language's input mapping or that are unicode letters, numbers and diacritics.

"""

import re
from typing import List

from g2p.exceptions import MappingMissing
from g2p.log import LOGGER
from g2p.mappings import Mapping, utils
from g2p.mappings.langs import LANGS_NETWORK
from g2p.mappings.utils import is_ipa
from g2p.shared_types import BaseTokenizer, Token


class Tokenizer(BaseTokenizer):
    """Base class for all g2p tokenizers; implements the default tokenizing behaviour.

    By default, a token is defined as a sequence of letters, numbers and
    diacritics, as defined in the Unicode Standard.
    """

    def __init__(self):
        self.inventory = []
        self.delim = ""
        self.case_sensitive = False
        # Hack for Tlingit where . is a letter when not word final:
        self.dot_is_letter = False
        self.escape_special = False

    def tokenize_aux(self, text):
        return text

    def is_word_character(self, characters):
        if not self.case_sensitive:
            characters = characters.lower()
        inventory_check = re.escape(characters) if self.escape_special else characters
        if inventory_check in self.inventory:
            return True
        if self.delim and characters == self.delim:
            return True
        return all(
            utils.get_unicode_category(c) in ["letter", "number", "diacritic"]
            for c in characters
        )

    def tokenize_text(self, text: str) -> List[Token]:
        matches = self.tokenize_aux(text)
        units = [Token(m, self.is_word_character(m)) for m in matches]
        if self.dot_is_letter:
            for i, unit in enumerate(units):
                if unit.text == "." and i + 1 < len(units) and units[i + 1].is_word:
                    unit.is_word = True
        return utils.merge_same_type_tokens(units)


class SpecializedTokenizer(Tokenizer):
    def __init__(self, mapping: Mapping):
        self.delim = ""
        self.inventory = mapping.inventory("in")
        self.lang = mapping.language_name
        self.case_sensitive = mapping.case_sensitive
        self.dot_is_letter = False
        self.escape_special = mapping.escape_special
        # create regex
        self._build_regex()

    def _build_regex(self):
        if not self.case_sensitive:
            self.inventory = [c.lower() for c in self.inventory]
        # Remove the indices, they're not part of the text input for the rules
        self.inventory = [re.sub(r"{[0-9]+}", "", x) for x in self.inventory]
        # Rules with "in": "è|é" have two distinct inputs, split them.
        self.inventory = [
            part
            for rule_input in self.inventory
            for part in re.split(r"(?<!\\)\|", rule_input)
        ]
        regex_pieces = sorted(self.inventory, key=lambda s: -len(s))
        regex_pieces = [re.escape(p) for p in regex_pieces]
        if self.delim:
            regex_pieces.append(self.delim)
        pattern = "|".join(regex_pieces + ["."])
        pattern = "(" + pattern + ")"
        # LOGGER.warning(f"pattern for {self.lang}: {pattern}.")
        flags = re.DOTALL
        if not self.case_sensitive:
            flags |= re.I
        self.regex = re.compile(pattern, flags)

    def tokenize_aux(self, text):
        return self.regex.findall(text)


class LexiconTokenizer(Tokenizer):
    """Lexicon-based tokenizer will consider any entry in the lexicon a token,
    even if it contains punctuation characters. For text not in the lexicon,
    falls back to the default tokenization.
    """

    def __init__(self, mapping: Mapping):
        super().__init__()
        self.mapping = mapping
        self.lang = mapping.language_name

    def _recursive_helper(self, tokens: list, output_tokens: list):
        """Emit the longest prefix found in the lexicon, if any, as a token.
        If None, emit the first unit as a token.
        Recursively process the rest of the units.
        """
        if not tokens:
            return
        if len(tokens) == 1:
            output_tokens.append(tokens[0])
            return
        for i in range(len(tokens), 0, -1):
            candidate = "".join([u.text for u in tokens[:i]])
            if utils.find_alignment(self.mapping.alignments, candidate.lower()):
                output_tokens.append(Token(candidate, True))
                return self._recursive_helper(tokens[i:], output_tokens)
        # No prefix found, emit the first unit as a token
        output_tokens.append(tokens[0])
        self._recursive_helper(tokens[1:], output_tokens)

    def tokenize_text(self, text: str) -> List[Token]:
        blocks = re.split(r"(\s+)", text)
        output_tokens = []
        for i, block in enumerate(blocks):
            if i % 2 == 1 and block:
                output_tokens.append(Token(block, False))
            else:
                default_tokens = super().tokenize_text(block)
                # Split non-word tokens into smaller parts for lexicon lookup
                candidate_tokens = utils.split_non_word_tokens(default_tokens)
                self._recursive_helper(candidate_tokens, output_tokens)

        return utils.merge_non_word_tokens(output_tokens)


class MultiHopTokenizer(SpecializedTokenizer):
    def __init__(self, mappings: List[Mapping]):
        self.delim = ""
        assert mappings
        self.inventory = sum([m.inventory("in") for m in mappings], [])
        self.lang = mappings[0].language_name
        self.case_sensitive = mappings[0].case_sensitive
        self.dot_is_letter = False
        self.escape_special = mappings[0].escape_special
        self._build_regex()
        # LOGGER.warning(pprint.pformat([self.lang, self.delim, self.case_sensitive, self.inventory]))


class TokenizerLibrary:
    """
    The TokenizerLibrary holds the collection of tokenizers that have been
    initialized so far. We don't initialize them all since that takes costly
    loading time; instead, we initialize each only as requested.
    """

    def __init__(self):
        self.tokenizers = {None: Tokenizer()}

    def make_tokenizer_key(self, in_lang, out_lang=None, tok_path=None):
        if not in_lang:
            return None
        if tok_path:
            return in_lang + "+tok_path=" + "-to-".join(tok_path)
        if not out_lang:
            out_lang = in_lang + "-ipa"
        return in_lang + "-to-" + out_lang

    def make_tokenizer(  # noqa C901
        self, in_lang, out_lang=None, tok_path=None
    ) -> Tokenizer:
        tokenizer_key = self.make_tokenizer_key(in_lang, out_lang, tok_path)
        if not self.tokenizers.get(tokenizer_key):
            # This tokenizer was not created yet, initialize it now.
            if tok_path:
                # LOGGER.warning(f"in_lang={in_lang} tok_path={tok_path}")
                if tok_path[0] != in_lang:
                    raise ValueError(
                        "calling make_tokenizer() with tok_path requires that tok_path[0] == in_lang"
                    )
                assert len(tok_path) >= 2
                if len(tok_path) == 2 or is_ipa(tok_path[1]):
                    out_lang = tok_path[1]
                elif len(tok_path) == 3 or is_ipa(tok_path[2]):
                    out_lang = tok_path[1:3]
                elif len(tok_path) > 3 and is_ipa(tok_path[3]):
                    out_lang = tok_path[1:4]
                else:
                    out_lang = tok_path[1:3]
            if not out_lang:
                try:
                    successors = list(LANGS_NETWORK.successors(in_lang))
                except KeyError:
                    successors = []
                ipa_successors = [x for x in successors if is_ipa(x)]
                # LOGGER.warning(pprint.pformat([in_lang, "->", successors, ipa_successors]))
                if ipa_successors:
                    # in_lang has an ipa successor, tokenize using it
                    # there currently are no langs with more than 1 IPA successor, but to
                    # be future-proof we'll arbitrarily take the first if there are more.
                    out_lang = ipa_successors[0]
                else:
                    # There is no direct IPA successor, look for a two-hop path to -ipa
                    for x in successors:
                        ipa_successors_two_hops = [
                            y for y in LANGS_NETWORK.successors(x) if is_ipa(y)
                        ]
                        # LOGGER.warning(pprint.pformat([in_lang, x, "->", [ipa_successors_two_hops]]))
                        if ipa_successors_two_hops:
                            out_lang = [x, ipa_successors_two_hops[0]]
                        break
                    # There is no two-hop IPA successor, use the first direct successor
                    if out_lang is None and successors:
                        out_lang = successors[0]
            # LOGGER.warning(f"Tokenizer for {in_lang} is {out_lang}.")
            if out_lang is None:
                # Default tokenizer:
                self.tokenizers[tokenizer_key] = self.tokenizers[None]
            elif isinstance(out_lang, list):
                # Build a multi-hop tokenizer
                assert len(out_lang) > 1
                try:
                    mappings = [
                        Mapping.find_mapping(in_lang=in_lang, out_lang=out_lang[0])
                    ]
                    for i in range(1, len(out_lang)):
                        mappings.append(
                            Mapping.find_mapping(
                                in_lang=out_lang[i - 1], out_lang=out_lang[i]
                            )
                        )
                    self.tokenizers[tokenizer_key] = MultiHopTokenizer(mappings)
                except MappingMissing:
                    self.tokenizers[tokenizer_key] = self.tokenizers[None]
                    LOGGER.warning(
                        f"missing mapping yet we looked for mappings in graph for {in_lang}-{out_lang}."
                    )
            else:
                # Build a one-hop tokenizer
                try:
                    mapping = Mapping.find_mapping(in_lang=in_lang, out_lang=out_lang)
                    if mapping.type == utils.MAPPING_TYPE.lexicon:
                        self.tokenizers[tokenizer_key] = LexiconTokenizer(mapping)
                    else:
                        self.tokenizers[tokenizer_key] = SpecializedTokenizer(mapping)
                except MappingMissing:
                    self.tokenizers[tokenizer_key] = self.tokenizers[None]
                    LOGGER.warning(
                        f"Cannot find mapping from '{in_lang}' to '{out_lang}'. Using default tokenizer instead"
                    )

            # Hack for Tlingit using dot as a letter when non word-final
            # We also apply it for sal-apa so it can work with Secwempetsín
            if in_lang == "tli" or in_lang == "sal-apa":
                self.tokenizers[tokenizer_key].dot_is_letter = True

        return self.tokenizers[tokenizer_key]


_the_tokenizer_library = TokenizerLibrary()


def make_tokenizer(in_lang=None, out_lang=None, tok_path=None) -> Tokenizer:
    """Make the tokenizer for input in language in_lang

    Logic used when only in_lang is provided:
    - if in_lang -> in_lang-ipa, or in_lang -> X-ipa exists, tokenize using the input
      inventory of that mapping.
    - elif in_lang -> X -> Y-ipa exists, e.g., tce -> tce-equiv -> tce-ipa, tokenize
      using the input inventory of those two hops.
    - otherwise, just use the default tokenizer, which accepts as part of words all
      unicode letter, numbers and diacritics

    Logic used when in_lang and out_lang are provided:
    - if in_lang -> out_lang exists, tokenize using the input inventory of that mapping
    - otherwise use the default tokenizer

    Logic used when in_lang and tok_path are provided:
    - use the first one or two hops in path, stopping at the first -ipa node
    """
    return _the_tokenizer_library.make_tokenizer(in_lang, out_lang, tok_path)


_deprecated_warning_printed = False


def get_tokenizer(*args, **kwargs):
    """Deprecated; use make_tokenizer() instead."""

    global _deprecated_warning_printed
    if not _deprecated_warning_printed:
        LOGGER.warning(
            "g2p.get_tokenizer() / g2p.mappings.tokenizer.get_tokenizer() is deprecated. Import and use g2p.make_tokenizer() instead."
        )
        _deprecated_warning_printed = True

    return make_tokenizer(*args, **kwargs)



#!/usr/bin/env python

"""Convert CMUDict to IPA, removing stress markers. We can only handle
one pronunciation at a time so just take the first one."""

import fileinput
import json
import re

with open("eng_arpabet_to_ipa.json") as f:
    mappings = json.load(f)
    ipa_map = dict((e["in"], e["out"]) for e in mappings)

comment_re = re.compile(r"#.*$")
entry_re = re.compile(r"^(\S+?)(\(\d+\))?\s+(.*)$")
stress_re = re.compile(r"\d+$")
for spam in fileinput.input():
    m = entry_re.match(comment_re.sub("", spam.strip()))
    if m is None:
        continue
    word, alt, phones = m.groups()
    if alt is not None:  # skip alterantes
        continue
    phones = "".join(
        ipa_map[np] for np in (stress_re.sub("", p) for p in phones.split())
    )
    print("\t".join((word, phones)))



#!/usr/bin/env python

"""
Test that the --config switch to g2p convert works

This test modifies the g2p database in memory, so it's important that it get
run last, in order to avoid changing the results of other unit test cases.

This file is also the right place to include other tests that make use of --config
to exercise different things.

We accomplish that in two ways:
 - in "./run.py dev", it gets appended after all the other tests, explicitly.
 - the file name has a "z" so it sorts last when "./run.py all" uses LOADER.discover()
"""

import json
import os
import tempfile
from pathlib import Path
from unittest import TestCase, main

import yaml
from click.testing import CliRunner

from g2p import exceptions
from g2p.cli import convert, generate_mapping
from g2p.mappings import Mapping
from g2p.mappings.utils import normalize
from g2p.tests.public import PUBLIC_DIR


class LocalConfigTest(TestCase):
    def setUp(self):
        self.runner = CliRunner()
        self.mappings_dir = Path(PUBLIC_DIR) / "mappings"

    def test_local_config(self) -> None:
        config_path = str(self.mappings_dir / "test.yaml")
        result = self.runner.invoke(
            convert,
            [
                "bbbb",
                "local-config-in",
                "local-config-out",
                "--config",
                config_path,
            ],
        )
        self.assertIn("aaaa", result.stdout)
        result = self.runner.invoke(
            convert,
            [
                "b",
                "local-config-in",
                "eng-ipa",
                "--config",
                config_path,
            ],
        )
        self.assertIn("ɑ", result.stdout)

    def test_case_insensitive_tokenizer(self) -> None:
        # Unit testing for https://github.com/ReadAlongs/Studio/issues/40
        # That issue was raised in Studio when the tokenizer was there, but
        # the tokenizer has been migrated to g2p since then, so the test and
        # fix belong here in g2p.

        # This test incidentally exercises passing --config a config file with
        # only one mapping in it, without the top-level "mappings:" list.
        tok_config = str(self.mappings_dir / "tokenize_punct_config-g2p.yaml")
        results = self.runner.invoke(
            convert, ["--tok", "--config", tok_config, "AAA-BBB", "tok-in", "tok-out"]
        )
        self.assertEqual(results.exit_code, 0)
        self.assertIn("aac_dbb", results.output)

        # While "AAA-BBB" gets tokenized as [word("AAA-BBB")], since "A-B" is
        # in the inventory, "D-C" gets tokenized as [word("D"), non-word("-"),
        # word("C")]. With rule D,d_end,,$, D will only become "d_end" if D is
        # at the end, which only occurs where we tokenize.
        results = self.runner.invoke(
            convert, ["--tok", "--config", tok_config, "D-C", "tok-in", "tok-out"]
        )
        self.assertEqual(results.exit_code, 0)
        self.assertIn("d_end-c", results.output)

    def test_null_mapping(self) -> None:
        """Empty lines in a mapping should just get ignored"""
        # Unit test case for bug fix: as of 2021-12-01, an empty rule would
        # cause the next rule to not get its match pattern created, and raise
        # an exception later. The null.csv mapping has such an empty rule,
        # which should get ignored, with the next rule, d->e, still working.
        null_config = str(self.mappings_dir / "null_config-g2p.yaml")
        results = self.runner.invoke(
            convert, ["--config", null_config, "x-ad-x", "null-in", "null-out"]
        )
        self.assertEqual(results.exit_code, 0)
        self.assertIn("x-be-x", results.output)

    def test_case_feeding_mapping(self) -> None:
        """Exercise the mapping using case to prevent feeding on in/out but not context"""
        case_feeding_config = str(self.mappings_dir / "case-feed" / "config-g2p.yaml")
        results = self.runner.invoke(
            convert,
            [
                "--config",
                case_feeding_config,
                "--tok",
                "ka-intinatin",
                "cf-in",
                "cf-out",
            ],
        )
        # print(results.output)
        self.assertEqual(results.exit_code, 0)
        self.assertIn("ke-antinetin", results.output)

    def test_missing_files(self) -> None:
        """Nice error messages when the mapping file or abbreviations file are missing"""
        with tempfile.TemporaryDirectory() as tmpdir:
            config_file = os.path.join(tmpdir, "mapping-file-not-found.yaml")
            with open(config_file, "wt", encoding="utf8") as f:
                yaml.dump({"mappings": [{"rules_path": "no-such-file.csv"}]}, f)
            results = self.runner.invoke(
                convert, ["--config", config_file, "a", "b", "c"]
            )
            self.assertNotEqual(results.exit_code, 0)
            self.assertIn(
                "No such file or directory", results.output + str(results.exception)
            )

        with tempfile.TemporaryDirectory() as tmpdir:
            config_file = os.path.join(tmpdir, "abbrev-file-not-found.yaml")
            with open(os.path.join(tmpdir, "tiny.csv"), "wt", encoding="utf8") as f:
                f.write("a,b\n")
            with open(config_file, "wt", encoding="utf8") as f:
                yaml.dump(
                    {
                        "mappings": [
                            {
                                "rules_path": "tiny.csv",
                                "abbreviations_path": "no-such-file.csv",
                            }
                        ]
                    },
                    f,
                )
            results = self.runner.invoke(
                convert, ["--config", config_file, "a", "b", "c"]
            )
            self.assertNotEqual(results.exit_code, 0)
            self.assertIn(
                "No such file or directory",
                results.output + str(results.exception),
            )

    def test_empty_rules(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            config_file = os.path.join(tmpdir, "empty-rules-file.yaml")
            with open(os.path.join(tmpdir, "empty.csv"), "wt", encoding="utf8") as f:
                pass
            with open(config_file, "wt", encoding="utf8") as f:
                yaml.dump({"mappings": [{"rules_path": "empty.csv"}]}, f)
            with self.assertRaises(exceptions.MalformedMapping) as e:
                # This is a deep pydantic exception, we should raise MalformedMapping
                Mapping.load_mapping_from_path(config_file)
            self.assertIn("empty.csv does not contain any rules", str(e.exception))
            results = self.runner.invoke(
                convert, ["--config", config_file, "a", "b", "c"]
            )
            self.assertNotEqual(results.exit_code, 0)
            self.assertIn(
                "empty.csv does not contain any rules",
                results.output + str(results.exception),
            )

    def test_generate_mapping(self) -> None:
        """Use a local config to test generate mapping with --from and --to"""
        # This test is rather hacky, because it relies on the fact that calling
        # g2p convert --config loads a config and keeps it in memory for the rest
        # of the unit tests. While that's a potential bug for testing g2p update,
        # and the very reason this file is called test_z_..., forcing it to be
        # called last, we're using it as a feature here.

        # It would be better to have a proper function to load in additional
        # mappings from a given config file, but right now that's not
        # implemented, and I don't want to make that a requirement for testing
        # "g2p generate-mapping --from/--to", so it'll have to wait.
        # TODO: write a internal load_from_config() function, or some such, and
        # factor out the repeated code to use it.

        # This first case has the side effect of loading gen-map_config-g2p.yaml
        config_path = str(self.mappings_dir / "gen-map_config-g2p.yaml")
        result = self.runner.invoke(
            convert, ["uyoesnmklbdt", "gm2", "gm2-ipa", "--config", config_path]
        )
        self.assertIn("uyɔɛsnmklbdt", result.stdout)
        # This second case confirms that gen-map_config-g2p.yaml is still loaded
        result = self.runner.invoke(convert, ["uyoesnmklbdt", "gm3a", "gm3-ipa"])
        self.assertIn("uyoesnmklbdt", result.stdout)

        # Now we do the real tests
        with tempfile.TemporaryDirectory() as output_dir_s:
            output_dir = Path(output_dir_s)
            # for debugging:
            # output_dir = Path(".") / "gen-map-tests"
            # output_dir.mkdir(exist_ok=True)
            # 1 mapping in to 1 mapping out
            result = self.runner.invoke(
                generate_mapping,
                ["--from", "gm1", "--to", "gm2", "--out-dir", output_dir_s],
            )
            self.assertEqual(result.exit_code, 0)
            with open(self.mappings_dir / "gm1-ipa_to_gm2-ipa.json", "r") as f:
                ref = json.load(f)
            with open(output_dir / "gm1-ipa_to_gm2-ipa.json", "r") as f:
                output = json.load(f)
            self.assertEqual(output, ref)

            # 2 mappings in to 1 mapping out
            result = self.runner.invoke(
                generate_mapping,
                ["--from", "gm3", "--to", "gm2", "--out-dir", output_dir_s],
            )
            self.assertEqual(result.exit_code, 0)
            with open(self.mappings_dir / "gm3-ipa_to_gm2-ipa.json", "r") as f:
                ref = json.load(f)
            with open(output_dir / "gm3-ipa_to_gm2-ipa.json", "r") as f:
                output = json.load(f)
            self.assertEqual(output, ref)

            # 1 mapping in to 2 mappings out
            result = self.runner.invoke(
                generate_mapping,
                ["--from", "gm2", "--to", "gm3", "--out-dir", output_dir_s],
            )
            self.assertEqual(result.exit_code, 0)
            with open(self.mappings_dir / "gm2-ipa_to_gm3-ipa.json", "r") as f:
                ref = json.load(f)
            with open(output_dir / "gm2-ipa_to_gm3-ipa.json", "r") as f:
                output = json.load(f)
            self.assertEqual(output, ref)

    def test_compose_NFC_NFD(self) -> None:
        config_path = str(self.mappings_dir / "compose.yaml")
        result = self.runner.invoke(
            convert,
            [
                normalize("é", "NFD"),
                "c1",
                "c3",
                "--no-tok",
                "--config",
                config_path,
                "-d",
                "-e",
            ],
        )
        self.assertEqual(result.exit_code, 0)
        self.assertIn("[[(0, 0), (1, 0)], [(0, 0), (0, 1)]]", result.output)
        self.assertIn(
            "[[('e', 'ò'), ('́', 'ò')], [('ò', 'u'), ('ò', '̀')]]", result.output
        )

        result = self.runner.invoke(
            convert,
            [
                normalize("é", "NFC"),
                "c1",
                "c3",
                "--no-tok",
                "--config",
                config_path,
                "-d",
                "-e",
            ],
        )
        self.assertEqual(result.exit_code, 0)
        self.assertIn("[[(0, 0)], [(0, 0), (0, 1)]]", result.output)
        self.assertIn("[[('é', 'ò')], [('ò', 'u'), ('ò', '̀')]]", result.output)

    def test_nofeed_indices(self) -> None:
        config_path = str(self.mappings_dir / "nofeed-indices.yaml")
        args = ("nofeed-indices-in", "nofeed-indices-out")
        result = self.runner.invoke(convert, ["ab", *args, "--config", config_path])
        self.assertIn("ced", result.stdout)
        result = self.runner.invoke(convert, ["abft", *args, "-d"])
        self.assertIn("cedft", result.stdout)
        result = self.runner.invoke(convert, ["deft", *args, "-d"])
        self.assertIn("ghit", result.stdout)
        result = self.runner.invoke(convert, ["aātaāabtaā", *args, "-e"])
        self.assertIn("aʼataʼacedtaʼa", result.stdout)

    def test_invalid_abbrev_path(self) -> None:
        config = """
            mappings:
             - name: "invalid"
               abbreviations_path: {}
               rules: []
        """
        with tempfile.TemporaryDirectory() as tmpdir_s:
            tmpdir = Path(tmpdir_s)
            # type bool is not valid
            with open(tmpdir / "invalid.yaml", "w", encoding="utf8") as fh:
                fh.write(config.format("true"))
            with self.assertRaises(Exception) as e:
                # Currently raise a deep pydantic exception, exceptions.MalformedMapping would be better
                Mapping.load_mapping_from_path(tmpdir / "invalid.yaml")
            message = str(e.exception)
            self.assertRegex(
                message, r"(?s)abbreviations_path.*Input is not a valid path"
            )
            # This fails because we don't raise MalformedMapping
            # self.assertRegex(
            #     message, r"Problem in config file:.*invalid.yaml"
            # )

            # type float is not valid
            with open(tmpdir / "invalid.yaml", "w", encoding="utf8") as fh:
                fh.write(config.format("1.0"))
            with self.assertRaises(Exception):
                # Currently raise a deep pydantic exception, exceptions.MalformedMapping would be better
                Mapping.load_mapping_from_path(tmpdir / "invalid.yaml")
            message = str(e.exception)
            self.assertRegex(
                message, r"(?s)abbreviations_path.*Input is not a valid path"
            )
            # This fails because we don't raise MalformedMapping
            # self.assertRegex(
            #     message, r"Problem in config file:.*invalid.yaml"
            # )

            # non-existent file
            with open(tmpdir / "invalid.yaml", "w", encoding="utf8") as fh:
                fh.write(config.format("file_not_found.csv"))
            # This is super unfriendly to the user, but that's what happens now.
            with self.assertRaises(FileNotFoundError):
                Mapping.load_mapping_from_path(tmpdir / "invalid.yaml")


if __name__ == "__main__":
    main()



#!/usr/bin/env python

""" Checks all data resources give 200s
"""

import json
import os
import re
from typing import Dict, Union
from unittest import TestCase, main

from fastapi.testclient import TestClient

from g2p.app import APP
from g2p.log import LOGGER
from g2p.mappings.langs import LANGS_NETWORK
from g2p.tests.public import __file__ as PUB_FILE

PUB_DIR = os.path.dirname(PUB_FILE)


class ResourceIntegrationTest(TestCase):
    """
    This tests that the api returns 200s for all basic
    GET requests.
    """

    def setUp(self):
        # routes
        self.conversion_route = "/g2p"
        for route in APP.routes:
            if route.path == "/api/v1":  # type: ignore
                self.api = route.app  # type: ignore
                self.routes = route.routes  # type: ignore
        self.routes_no_args = [
            route.path
            for route in self.routes
            if "{" not in route.path and route.path != self.conversion_route
        ]
        self.routes_only_args = [
            route.path for route in self.routes if "{" in route.path
        ]
        self.client = TestClient(self.api)
        # args
        self.arg_match = re.compile(r"\{[a-z:]+\}")
        self.args_to_check = "node"

    def test_response_code(self):
        """
        Ensure all routes return 200
        """
        for rt in self.routes_no_args:
            try:
                with self.assertLogs():  # silence the logs by asserting them
                    r = self.client.get(rt)
                self.assertEqual(r.status_code, 200)
                LOGGER.debug("Route " + rt + " returned " + str(r.status_code))
            except Exception as exc:
                LOGGER.error("Couldn't connect. Is the API running? %s", exc)

    def test_response_code_with_args(self):
        """
        Ensure all args return 200
        """
        for ep in self.routes_only_args:
            for node in LANGS_NETWORK.nodes:
                rt = re.sub(self.arg_match, node, ep)
                try:
                    with self.assertLogs():  # silence the logs by asseting them
                        r = self.client.get(rt)
                    self.assertEqual(r.status_code, 200)
                except Exception as exc:
                    LOGGER.error("Couldn't connect. Is the API running? %s", exc)
            LOGGER.debug(
                "Successfully tested "
                + str(len(LANGS_NETWORK.nodes))
                + " node resources at route "
                + ep
                + " ."
            )

    def test_g2p_conversion(self):
        """
        Ensure conversion returns proper response
        """
        params: Dict[str, Union[str, bool]] = {
            "in-lang": "dan",
            "out-lang": "eng-arpabet",
            "text": "hej",
            "debugger": True,
            "index": True,
        }
        minimal_params: Dict[str, Union[str, bool]] = {
            "in-lang": "dan",
            "out-lang": "eng-arpabet",
            "text": "hej",
            "debugger": False,
            "index": False,
        }
        bad_params = {"in-lang": "dan", "out-lang": "moh", "text": "hej"}
        same_params = {"in-lang": "dan", "out-lang": "dan", "text": "hej"}
        missing_params = {
            "in-lang": "not-here",
            "out-lang": "eng-arpabet",
            "text": "hej",
        }
        self.maxDiff = None
        with self.assertLogs():
            response = self.client.get(self.conversion_route, params=params)  # type: ignore
        res_json = response.json()
        self.assertEqual(response.status_code, 200)
        with open(os.path.join(PUB_DIR, "sample_response.json")) as f:
            data = json.load(f)
        self.assertEqual(res_json, data)
        # check minimal response
        with self.assertLogs():
            minimal_response = self.client.get(
                self.conversion_route, params=minimal_params
            )
        data["debugger"] = False
        data["index"] = False
        self.assertEqual(minimal_response.status_code, 200)
        self.assertEqual(minimal_response.json(), data)
        with self.assertLogs(LOGGER, level="ERROR"):
            bad_response = self.client.get(self.conversion_route, params=bad_params)
        with self.assertLogs(LOGGER, level="ERROR"):
            same_response = self.client.get(self.conversion_route, params=same_params)
        self.assertEqual(bad_response.status_code, 400)
        self.assertEqual(same_response.status_code, 400)
        with self.assertLogs(LOGGER, level="ERROR"):
            missing_response = self.client.get(
                self.conversion_route, params=missing_params
            )
        self.assertEqual(missing_response.status_code, 404)
        invalid_params = {
            "in-lang": "dan",
            "out-lang": "eng-arpabet",
            "text": "hej",
            "debugger": "THIS IS NOT A BOOLEAN!!!",
            "index": "NEITHER IS THIS!!!",
        }
        with self.assertLogs(LOGGER, level="ERROR"):
            invalid_response = self.client.get(
                self.conversion_route, params=invalid_params
            )
        self.assertEqual(invalid_response.status_code, 422)

    def test_g2p_conversion_with_tok(self):
        params_with_tok: Dict[str, Union[str, bool]] = {
            "in-lang": "fra",
            "out-lang": "eng-arpabet",
            "text": "ceci, celà",
            "debugger": True,
            "index": True,
            "tokenize": True,
        }
        with self.assertLogs():
            response = self.client.get(self.conversion_route, params=params_with_tok)
        self.assertEqual(response.status_code, 200)
        res_json_tok = response.json()
        self.assertEqual(res_json_tok["debugger"][0][0][0]["input"], "ceci")

        params_no_tok: Dict[str, Union[str, bool]] = {
            "in-lang": "fra",
            "out-lang": "eng-arpabet",
            "text": "ceci, celà",
            "debugger": True,
            "index": True,
            "tokenize": False,
        }
        with self.assertLogs():
            response = self.client.get(self.conversion_route, params=params_no_tok)
        self.assertEqual(response.status_code, 200)
        res_json_no_tok = response.json()
        self.assertNotEqual(res_json_tok, res_json_no_tok)
        self.assertEqual(res_json_no_tok["debugger"][0][0][0]["input"], "ceci, celà")

        self.assertNotEqual(res_json_tok["debugger"], res_json_no_tok["debugger"])


if __name__ == "__main__":
    main()



"""
Common code to access the data in tests/public/data

Test files (in g2p/tests/public/data) are either .csv, .psv, or .tsv files, the
only difference being the delimiter used (comma, pipe (|), or tab).

Each line in the test file consists of SOURCE,TARGET,INPUT,OUTPUT
"""

import csv
import os
from glob import glob
from typing import List

from g2p.log import LOGGER

"""Directory where the public test data is located"""
DATA_DIR = os.path.dirname(__file__)

"""Cached for langs data, so we don't have to reload it multiple times"""
loaded_langs_to_test = None


def read_data_file(fn, delimiter):
    rows = []
    with open(fn, encoding="utf-8") as csvfile:
        reader = csv.reader(csvfile, delimiter=delimiter)
        for i, row in enumerate(reader):
            if len(row) == 0 or (len(row) == 1 and row[0].strip() == ""):
                # skip empty and comment lines
                continue
            elif row[0][:1] == "#":
                # skip comments, but check them for stray quotes
                if "\n" in "".join(row):
                    raise Exception(
                        f"Comment on line {i+1} of {fn} absorbed the next data line(s)."
                        "Please remove stray double quotes."
                    )
                continue
            elif len(row) < 4:
                LOGGER.warning(
                    f"Row in {fn} containing values {row} does not have the right values."
                    f"Please check your data."
                )
            else:
                row.append(f"{fn}:{i+1}")
                rows.append(row)
    return rows


def load_neural_test_data() -> List[List[str]]:
    """Load neural data for test data

    Returns:
        List[List[in_lang, out_lang, in_text, out_text, filename:lineno]]
    """
    return read_data_file(os.path.join(DATA_DIR, "neural.psv"), "|")


def load_public_test_data() -> List[List[str]]:
    """Load public/data/*.?sv for test data in various languages

    Returns: List[List[in_lang, out_lang, in_text, out_text, filename:lineno]]
    """
    global loaded_langs_to_test
    if loaded_langs_to_test is not None:
        return loaded_langs_to_test

    langs_to_test = []
    for fn in sorted(glob(os.path.join(DATA_DIR, "*.*sv"))):
        if "neural" in fn:
            continue
        if fn.endswith("csv"):
            delimiter = ","
        elif fn.endswith("psv"):
            delimiter = "|"
        elif fn.endswith("tsv"):
            delimiter = "\t"
        langs_to_test += read_data_file(fn, delimiter)

    loaded_langs_to_test = langs_to_test
    return langs_to_test



#!/usr/bin/env python

"""
    Unittests for index preservation
"""

from unicodedata import normalize
from unittest import TestCase, main

from g2p.log import LOGGER
from g2p.mappings import Mapping
from g2p.transducer import Transducer


class IndicesTest(TestCase):
    """Basic Transducer Test
    Preserve character-level mappings:

    Test Case #1
        # Simple conversion

        0 1 2 3
        t e s t

        p e s t
        0 1 2 3

        [ ((0, 't'), (0, 'p')),
          ((1, 'e'), (1, 'e')),
          ((2, 's'), (2, 's')),
          ((3, 't'), (3, 't')) ]

    Test Case #2:
        # Allow for deletion of segments

        0 1 2 3
        t e s t

        t s t
        0 1 2

        [ ((0, 't'), (0, 't')),
          ((1, 'e'), (1, '')),
          ((2, 's'), (2, 's')),
          ((3, 't'), (3, 't')) ]

    Test Case #3
        # Allow for one-to-many

        0 1 2 3
        t e s t

        c h e s t
        0 1 2 3 4

        [ ((0, 't'), (0, 'c')),
          ((0, 't'), (1, 'h')),
          ((1, 'e'), (2, 'e')),
          ((2, 's'), (3, 's')),
          ((3, 't'), (4, 't')) ]

    Test Case #4
        # Allow for many-to-one

        0 1 2 3
        t e s t

        p s t
        0 1 2

        [ ((0, 't'), (0, 'p')),
          ((1, 'e'), (0, 'p')),
          ((2, 's'), (1, 's')),
          ((3, 't'), (2, 't')) ]

    Test Case #5
        # Allow for epenthesis
         0 1 2 3
         t e s t

         t e s t y
         0 1 2 3 4

        [ ((-1, 'y'), (4, 'y')),
          ((0, 't'), (0, 't')),
          ((1, 'e'), (1, 'e')),
          ((2, 's'), (2, 's')),
          ((3, 't'), (3, 't')) ]

     Test Case #6
        # Allow metathesis
         0 1 2 3
         t e s t

         t s e t
         0 1 2 3

        [ ((0, 't'), (0, 't')),
          ((1, 'e'), (2, 'e')),
          ((2, 's'), (1, 's')),
          ((3, 't'), (3, 't')) ]

    Test Case #7
        # Allow order-sensitive operations
        0 1 2 3
        t e s t

        t e s h t
        0 1 2 3 4

        t e s t
        0 1 2 3

        AS IS

        [ ((0, 't'), (0, 't')),
          ((1, 'e'), (1, 'e')),
          ((2, 's'), (2, 's')),
          ((3, 't'), (3, 't')) ]

          or not

        [ ((0, 't'), (0, 't')),
          ((1, 'e'), (1, 'e')),
          ((2, 's'), (2, 's')),
          ((2, 's'), (3, 'h')),
          ((3, 't'), (4, 't')) ]

    Test Case #8
        # Allow multiple processes which alter the indices
        0 1 2 3
        t e s t

        c h e s t
        0 1 2 3 4

        c h e s s
        0 1 2 3 4

        [ ((0, 't'), (0, 'c')),
          ((1, 'e'), (1, 'h')),
          ((1, 'e'), (2, 'e')),
          ((2, 's'), (3, 's')),
          ((3, 't'), (4, 's')) ]

    Test Case # 9
        # Allow multiple character deletion
        0 1
        a a

        None None

        [ ((0, 'a'), (None, '')),
          ((1, 'a'), (None, '')) ]

    Test Case # 10
        # Another deletion test
        0 1 2
        a b c

        a
        0

        [ ((0, 'a'), (0, 'a')),
          ((1, 'b'), (0, '')),
          ((1, 'c'), (0, '')) ]

    Test case # 11
        # Sort of an insertion test (empty inputs are not allowed)

    Test case # 12
        # Verify that empty inputs are not allowed
    """

    def __init__(self, *args):
        # Let's use __init__() to set all these up just once at class creation
        # time, instead of setUp() which repeatedly does it for each test case
        super().__init__(*args)
        self.test_mapping_one = Mapping(
            rules=[{"in": "t", "out": "p", "context_after": "e"}]
        )
        self.test_mapping_two = Mapping(rules=[{"in": "e", "out": ""}])
        self.test_mapping_three = Mapping(
            rules=[{"in": "t", "out": "ch", "context_after": "e"}]
        )
        self.test_mapping_four = Mapping(rules=[{"in": "te", "out": "p"}])
        # We know this issues a warning, so let's silence it by asserting it.
        with self.assertLogs(LOGGER, "WARNING"):
            self.test_mapping_five = Mapping(
                rules=[
                    {"context_before": "t", "context_after": "$", "in": "", "out": "y"}
                ]
            )
        self.test_mapping_six = Mapping(rules=[{"in": "e{1}s{2}", "out": "s{2}e{1}"}])
        self.test_mapping_seven = Mapping(
            rules=[{"in": "s", "out": "sh"}, {"in": "sh", "out": "s"}],
            rule_ordering="apply-longest-first",
        )
        self.test_mapping_seven_as_written = Mapping(
            rules=[{"in": "s", "out": "sh"}, {"in": "sh", "out": "s"}]
        )
        self.test_mapping_eight = Mapping(
            rules=[{"in": "te", "out": "che"}, {"in": "t", "out": "s"}]
        )
        self.test_mapping_nine = Mapping(rules=[{"in": "aa", "out": ""}])
        self.test_mapping_ten = Mapping(rules=[{"in": "abc", "out": "a"}])
        self.test_mapping_eleven = Mapping(rules=[{"in": "a", "out": "aaaa"}])
        self.test_mapping_combining = Mapping(
            rules=[{"in": "k{1}\u0313{2}", "out": "'{2}k{1}"}]
        )
        self.test_mapping_wacky = Mapping(
            rules=[
                {
                    "in": "\U0001f600{1}\U0001f603\U0001f604{2}\U0001f604{3}",
                    "out": "\U0001f604\U0001f604\U0001f604{2}\U0001f604{3}\U0001f604{1}",
                }
            ]
        )
        self.test_mapping_wacky_lite = Mapping(
            rules=[{"in": "a{1}bc{2}c{3}", "out": "ccc{2}c{3}c{1}"}]
        )
        self.test_mapping_circum = Mapping(
            rules=[{"in": "a{1}c{2}", "out": "c{2}a{1}c{2}"}]
        )
        self.test_mapping_explicit_equal_1 = Mapping(
            rules=[{"in": "a{1}b{1}", "out": "c{1}d{1}"}]
        )
        self.test_mapping_explicit_equal_2 = Mapping(
            rules=[{"in": "ab{1}", "out": "cd{1}"}]
        )
        self.test_mapping_explicit_equal_3 = Mapping(rules=[{"in": "ab", "out": "cd"}])
        self.test_mapping_explicit_equal_4 = Mapping(
            rules=[{"in": "a{1}b{2}", "out": "c{1}d{2}"}]
        )
        self.test_issue_173_1 = Mapping(
            rules=[
                {"in": "x{1}y{2}z{3}", "out": "a{2}b{1}"},
                {"in": "d{1}e{2}f{3}", "out": "d{1}e{2}f{3}"},
            ]
        )
        self.test_issue_173_2 = Mapping(
            rules=[
                {"in": "x{1}y{2}z{3}", "out": "a{1}b{2}"},
                {"in": "d{1}e{2}f{3}", "out": "d{1}e{2}f{3}"},
            ]
        )
        self.test_issue_157_mapping = Mapping(
            rules=[
                {"in": "a", "out": "d"},
                {"in": "bc", "out": "e"},
                {"in": "g{1}h{2}i{3}", "out": "G{2}H{1}I{3}J{1}"},
                {"in": "m{1}n{2}", "out": "N{2}M{1}"},
            ]
        )
        self.test_feeding_mapping_1 = Mapping(
            rules=[{"in": "ab", "out": "a"}, {"in": "a", "out": "cd"}]
        )
        self.test_feeding_mapping_2 = Mapping(
            rules=[{"in": "a", "out": "cd"}, {"in": "cd", "out": "b"}]
        )
        self.test_issue_173_3 = Mapping(rules=[{"in": "ab{1}c{2}", "out": "X{1}Y{2}"}])
        self.test_issue_173_4 = Mapping(rules=[{"in": "a{1}bc{2}", "out": "xy{1}z{2}"}])
        self.trans_one = Transducer(self.test_mapping_one)
        self.trans_two = Transducer(self.test_mapping_two)
        self.trans_three = Transducer(self.test_mapping_three)
        self.trans_four = Transducer(self.test_mapping_four)
        self.trans_six = Transducer(self.test_mapping_six)
        self.trans_seven = Transducer(self.test_mapping_seven)
        self.test_seven_as_written = Transducer(self.test_mapping_seven_as_written)
        self.trans_eight = Transducer(self.test_mapping_eight)
        self.trans_nine = Transducer(self.test_mapping_nine)
        self.trans_ten = Transducer(self.test_mapping_ten)
        self.trans_eleven = Transducer(self.test_mapping_eleven)
        self.trans_combining = Transducer(self.test_mapping_combining)
        self.trans_wacky = Transducer(self.test_mapping_wacky)
        self.trans_wacky_lite = Transducer(self.test_mapping_wacky_lite)
        self.trans_circum = Transducer(self.test_mapping_circum)
        self.trans_explicit_equal_1 = Transducer(self.test_mapping_explicit_equal_1)
        self.trans_explicit_equal_2 = Transducer(self.test_mapping_explicit_equal_2)
        self.trans_explicit_equal_3 = Transducer(self.test_mapping_explicit_equal_3)
        self.trans_explicit_equal_4 = Transducer(self.test_mapping_explicit_equal_4)
        self.trans_173_1 = Transducer(self.test_issue_173_1)
        self.trans_173_2 = Transducer(self.test_issue_173_2)
        self.trans_173_3 = Transducer(self.test_issue_173_3)
        self.trans_173_4 = Transducer(self.test_issue_173_4)
        self.trans_157 = Transducer(self.test_issue_157_mapping)
        self.trans_feeding_1 = Transducer(self.test_feeding_mapping_1)
        self.trans_feeding_2 = Transducer(self.test_feeding_mapping_2)

    def test_feeding(self):
        """Test feeding"""
        transducer_1 = self.trans_feeding_1("ab")
        self.assertEqual(transducer_1.output_string, "cd")
        self.assertEqual(transducer_1.edges, [(0, 0), (0, 1), (1, 0), (1, 1)])
        # because of "crossed" indices, we get one single monotonic alignment
        self.assertEqual(transducer_1.substring_alignments(), [("ab", "cd")])
        transducer_2 = self.trans_feeding_2("a")
        self.assertEqual(transducer_2.output_string, "b")
        self.assertEqual(transducer_2.edges, [(0, 0)])
        self.assertEqual(transducer_2.substring_alignments(), [("a", "b")])

    def test_issue_157(self):
        """Test explicit problem from Issue 157"""
        transducer = self.trans_157("abcmn")
        self.assertEqual(transducer.output_string, "deNM")
        self.assertEqual(transducer.edges, [(0, 0), (1, 1), (2, 1), (3, 3), (4, 2)])
        self.assertEqual(
            transducer.substring_alignments(), [("a", "d"), ("bc", "e"), ("mn", "NM")]
        )

    def test_issue_173(self):
        """Test explicit problems from Issue 173"""
        transducer_1 = self.trans_173_1("xyzmndef")
        transducer_2 = self.trans_173_2("xyzmndef")
        transducer_3 = self.trans_173_3("abc")
        transducer_4 = self.trans_173_4("abc")
        self.assertEqual(transducer_1.output_string, "abmndef")
        self.assertEqual(transducer_2.output_string, "abmndef")
        self.assertEqual(transducer_3.output_string, "XY")
        self.assertEqual(transducer_4.output_string, "xyz")
        self.assertEqual(
            transducer_1.edges,
            [(0, 1), (1, 0), (2, 0), (3, 2), (4, 3), (5, 4), (6, 5), (7, 6)],
        )
        self.assertEqual(
            transducer_1.substring_alignments(),
            [("xyz", "ab"), ("m", "m"), ("n", "n"), ("d", "d"), ("e", "e"), ("f", "f")],
        )
        self.assertEqual(
            transducer_2.edges,
            [(0, 0), (1, 1), (2, 1), (3, 2), (4, 3), (5, 4), (6, 5), (7, 6)],
        )
        self.assertEqual(
            transducer_2.substring_alignments(),
            [
                ("x", "a"),
                ("yz", "b"),
                ("m", "m"),
                ("n", "n"),
                ("d", "d"),
                ("e", "e"),
                ("f", "f"),
            ],
        )
        self.assertEqual(transducer_3.edges, [(0, 0), (1, 0), (2, 1)])
        self.assertEqual(transducer_3.substring_alignments(), [("ab", "X"), ("c", "Y")])
        self.assertEqual(transducer_4.edges, [(0, 0), (0, 1), (1, 2), (2, 2)])
        self.assertEqual(
            transducer_4.substring_alignments(), [("a", "xy"), ("bc", "z")]
        )

    def test_explicit_equal(self):
        """Test synonymous syntax for explicit indices"""
        explicit_1 = self.trans_explicit_equal_1("ab")
        explicit_2 = self.trans_explicit_equal_2("ab")
        explicit_3 = self.trans_explicit_equal_4("ab")
        implicit = self.trans_explicit_equal_3("ab")
        self.assertEqual(explicit_1.output_string, "cd")
        self.assertEqual(explicit_2.output_string, "cd")
        self.assertEqual(implicit.output_string, "cd")
        self.assertEqual(explicit_3.output_string, "cd")
        self.assertEqual(explicit_1.edges, [(0, 0), (1, 1)])
        self.assertEqual(explicit_2.edges, [(0, 0), (1, 1)])
        self.assertEqual(implicit.edges, [(0, 0), (1, 1)])
        self.assertEqual(explicit_3.edges, [(0, 0), (1, 1)])

    def test_no_indices(self):
        """Test straightforward conversion without returning indices."""
        transducer = self.trans_combining("k\u0313am")
        self.assertEqual(transducer.output_string, "'kam")

    def test_combining(self):
        """Test index preserving combining characters"""
        transducer = self.trans_combining("k\u0313am")
        self.assertEqual(transducer.output_string, "'kam")
        self.assertEqual(transducer.edges, [(0, 1), (1, 0), (2, 2), (3, 3)])

    def test_wacky(self):
        """Test weird Unicode emoji transformation..."""
        transducer_lite = self.trans_wacky_lite("abcc")
        transducer_lite_extra = self.trans_wacky_lite("abcca")
        self.assertEqual(transducer_lite.output_string, "ccccc")
        self.assertEqual(transducer_lite_extra.output_string, "ccccca")
        self.assertEqual(
            transducer_lite.edges, [(0, 4), (1, 0), (2, 1), (2, 2), (3, 3)]
        )
        self.assertEqual(transducer_lite.substring_alignments(), [("abcc", "ccccc")])
        self.assertEqual(
            transducer_lite_extra.edges,
            [(0, 4), (1, 0), (2, 1), (2, 2), (3, 3), (4, 5)],
        )
        self.assertEqual(
            transducer_lite_extra.substring_alignments(),
            [("abcc", "ccccc"), ("a", "a")],
        )
        transducer_no_i = self.trans_wacky("\U0001f600\U0001f603\U0001f604\U0001f604")
        self.assertEqual(
            transducer_no_i.output_string,
            "\U0001f604\U0001f604\U0001f604\U0001f604\U0001f604",
        )
        transducer = self.trans_wacky("\U0001f600\U0001f603\U0001f604\U0001f604")
        self.assertEqual(
            transducer.output_string,
            "\U0001f604\U0001f604\U0001f604\U0001f604\U0001f604",
        )
        self.assertEqual(transducer.edges, [(0, 4), (1, 0), (2, 1), (2, 2), (3, 3)])
        self.assertEqual(
            transducer.substring_alignments(),
            [
                (
                    "\U0001f600\U0001f603\U0001f604\U0001f604",
                    "\U0001f604\U0001f604\U0001f604\U0001f604\U0001f604",
                )
            ],
        )

    def test_circum(self):
        """Test circumfixing"""
        transducer = self.trans_circum("ac")
        self.assertEqual(transducer.output_string, "cac")
        self.assertEqual(transducer.edges, [(0, 1), (1, 0), (1, 2)])
        self.assertEqual(transducer.substring_alignments(), [("ac", "cac")])

    def test_case_one(self):
        """Test case one"""
        transducer = self.trans_one("test")
        self.assertEqual(transducer.output_string, "pest")
        self.assertEqual(transducer.edges, [(0, 0), (1, 1), (2, 2), (3, 3)])
        self.assertEqual(
            transducer.substring_alignments(),
            [("t", "p"), ("e", "e"), ("s", "s"), ("t", "t")],
        )
        transducer = self.trans_one("")
        self.assertEqual(transducer.output_string, "")
        self.assertEqual(transducer.edges, [])
        self.assertEqual(transducer.substring_alignments(), [])

    def test_case_two(self):
        transducer = self.trans_two("test")
        self.assertEqual(transducer.output_string, "tst")
        self.assertEqual(transducer.edges, [(0, 0), (1, 0), (2, 1), (3, 2)])
        self.assertEqual(
            transducer.substring_alignments(), [("te", "t"), ("s", "s"), ("t", "t")]
        )

    def test_case_three(self):
        transducer = self.trans_three("test")
        self.assertEqual(transducer.output_string, "chest")
        self.assertEqual(transducer.edges, [(0, 0), (0, 1), (1, 2), (2, 3), (3, 4)])
        self.assertEqual(
            transducer.substring_alignments(),
            [("t", "ch"), ("e", "e"), ("s", "s"), ("t", "t")],
        )

    def test_case_four(self):
        transducer = self.trans_four("test")
        self.assertEqual(transducer.output_string, "pst")
        self.assertEqual(transducer.edges, [(0, 0), (1, 0), (2, 1), (3, 2)])
        self.assertEqual(
            transducer.substring_alignments(), [("te", "p"), ("s", "s"), ("t", "t")]
        )

    def test_case_six(self):
        transducer = self.trans_six("test")
        self.assertEqual(transducer.output_string, "tset")
        self.assertEqual(transducer.edges, [(0, 0), (1, 2), (2, 1), (3, 3)])
        self.assertEqual(
            transducer.substring_alignments(), [("t", "t"), ("es", "se"), ("t", "t")]
        )

    def test_case_long_six(self):
        transducer = self.trans_six("esesse")
        self.assertEqual(transducer.output_string, "sesese")
        # Ensure that *minimal* monotonic segments are output
        self.assertEqual(
            transducer.substring_alignments(),
            [("es", "se"), ("es", "se"), ("s", "s"), ("e", "e")],
        )

    def test_case_seven(self):
        transducer_as_written = self.test_seven_as_written("test")
        self.assertEqual(transducer_as_written.output_string, "test")
        self.assertEqual(transducer_as_written.edges, [(0, 0), (1, 1), (2, 2), (3, 3)])
        self.assertEqual(
            transducer_as_written.substring_alignments(),
            [("t", "t"), ("e", "e"), ("s", "s"), ("t", "t")],
        )
        transducer = self.trans_seven("test")
        self.assertEqual(transducer.output_string, "tesht")
        self.assertEqual(transducer.edges, [(0, 0), (1, 1), (2, 2), (2, 3), (3, 4)])
        self.assertEqual(
            transducer.substring_alignments(),
            [("t", "t"), ("e", "e"), ("s", "sh"), ("t", "t")],
        )

    def test_case_eight(self):
        transducer = self.trans_eight("test")
        self.assertEqual(transducer.output_string, "chess")
        self.assertEqual(transducer.edges, [(0, 0), (1, 1), (1, 2), (2, 3), (3, 4)])
        self.assertEqual(
            transducer.substring_alignments(),
            [("t", "c"), ("e", "he"), ("s", "s"), ("t", "s")],
        )

    def test_case_nine(self):
        transducer = self.trans_nine("aa")
        self.assertEqual(transducer.output_string, "")
        self.assertEqual(transducer.edges, [(0, None), (1, None)])
        # Support deletions in substring_alignments
        self.assertEqual(transducer.substring_alignments(), [("aa", "")])
        transducer = self.trans_nine("aabbaab")
        self.assertEqual(transducer.output_string, "bbb")
        self.assertEqual(
            transducer.edges,
            [(0, 0), (1, 0), (2, 0), (3, 1), (4, 1), (5, 1), (6, 2)],
        )
        # Support deletions in substring_alignments.  NOTE: these
        # alignments are quite bogus due to the ad-hoc treatment of
        # deletions by rule-based mappings
        self.assertEqual(
            transducer.substring_alignments(),
            [("aab", "b"), ("baa", "b"), ("b", "b")],
        )

    def test_case_ten(self):
        transducer = self.trans_ten("abc")
        self.assertEqual(transducer.output_string, "a")
        self.assertEqual(transducer.edges, [(0, 0), (1, 0), (2, 0)])
        self.assertEqual(transducer.substring_alignments(), [("abc", "a")])

    def test_case_eleven(self):
        transducer = self.trans_eleven("a")
        self.assertEqual(transducer.output_string, "aaaa")
        self.assertEqual(transducer.edges, [(0, 0), (0, 1), (0, 2), (0, 3)])
        self.assertEqual(transducer.substring_alignments(), [("a", "aaaa")])

    def test_case_twelve(self):
        # Empty inputs are not allowed (should it actually throw an exception?)
        with self.assertLogs(LOGGER) as cm:
            self.test_mapping_twelve = Mapping(
                rules=[{"in": "", "out": "aa", "context_before": "b"}]
            )
            self.trans_twelve = Transducer(self.test_mapping_twelve)
            transducer = self.trans_twelve("b")
        self.assertIn(
            "disallowed",
            cm.output[0],
            "it should warn that empty inputs are disallowed",
        )
        self.assertEqual(transducer.output_string, "b")

    def test_case_acdc(self):
        transducer = Transducer(
            Mapping(rules=[{"in": "a{1}c{2}", "out": "c{2}a{1}c{2}"}])
        )
        tg = transducer("acdc")
        self.assertEqual(tg.output_string, "cacdc")
        self.assertEqual(tg.edges, [(0, 1), (1, 0), (1, 2), (2, 3), (3, 4)])
        self.assertEqual(
            tg.substring_alignments(), [("ac", "cac"), ("d", "d"), ("c", "c")]
        )

    def test_case_acac(self):
        transducer = Transducer(Mapping(rules=[{"in": "ab{1}c{2}", "out": "ab{2}"}]))
        transducer_default = Transducer(
            Mapping(rules=[{"in": "ab", "out": ""}, {"in": "c", "out": "ab"}])
        )
        tg = transducer("abcabc")
        self.assertEqual(tg.output_string, "abab")
        self.assertEqual(
            tg.edges,
            [
                (0, 0),
                (1, 0),
                (2, 0),
                (2, 1),
                (3, 1),
                (4, 1),
                (5, 2),
                (5, 3),
            ],
        )
        self.assertEqual(tg.substring_alignments(), [("abcab", "ab"), ("c", "ab")])
        tg_default = transducer_default("abcabc")
        self.assertEqual(tg_default.output_string, "abab")
        self.assertEqual(
            tg_default.edges,
            [
                (0, 0),
                (1, 0),
                (2, 0),
                (2, 1),
                (3, 1),
                (4, 1),
                (5, 2),
                (5, 3),
            ],
        )
        self.assertEqual(
            tg_default.substring_alignments(), [("abcab", "ab"), ("c", "ab")]
        )

    def test_arpabet(self):
        transducer = Transducer(
            Mapping(
                rules=[{"in": "ĩ", "out": "IY N"}], norm_form="NFC", out_delimiter=" "
            )
        )
        transducer_nfd = Transducer(
            Mapping(
                rules=[{"in": "ĩ", "out": "IY N"}], norm_form="NFD", out_delimiter=" "
            )
        )
        tg = transducer(normalize("NFC", "ĩĩ"))
        tg_nfd = transducer_nfd(normalize("NFD", "ĩĩ"))
        self.assertEqual(tg.output_string, "IY N IY N ")
        self.assertEqual(tg_nfd.output_string, "IY N IY N ")
        self.assertEqual(
            tg.edges,
            [
                (0, 0),
                (0, 1),
                (0, 2),
                (0, 3),
                (0, 4),
                (1, 5),
                (1, 6),
                (1, 7),
                (1, 8),
                (1, 9),
            ],
        )
        self.assertEqual(tg.substring_alignments(), [("ĩ", "IY N "), ("ĩ", "IY N ")])
        self.assertEqual(
            tg_nfd.edges,
            [
                (0, 0),
                (1, 1),
                (1, 2),
                (1, 3),
                (1, 4),
                (2, 5),
                (3, 6),
                (3, 7),
                (3, 8),
                (3, 9),
            ],
        )
        self.assertEqual(
            tg_nfd.substring_alignments(),
            [("i", "I"), ("̃", "Y N "), ("i", "I"), ("̃", "Y N ")],
        )


if __name__ == "__main__":
    main()



"""
This file is for constants that can be initialized without any (expensive) dependencies.
"""

import os

DISTANCE_METRICS = [
    "weighted_feature_edit_distance",
    "hamming_feature_edit_distance",
    "feature_edit_distance",
    "dolgo_prime_distance",
    "fast_levenshtein_distance",
    "levenshtein_distance",
]

LANGS_DIR = os.path.join(os.path.dirname(__file__), "mappings", "langs")
LANGS_FILE_NAME = "langs.json.gz"
NETWORK_FILE_NAME = "network.json.gz"



"""

Module for all things related to lookup tables

"""

import csv
import json
import os
import re
from copy import deepcopy
from functools import partial
from pathlib import Path
from typing import Callable, Dict, List, Pattern, Union

import yaml
from pydantic import BaseModel

from g2p import exceptions
from g2p._version import version_tuple
from g2p.exceptions import NeuralDependencyError
from g2p.log import LOGGER
from g2p.mappings.langs import _LANGS, _MAPPINGS_AVAILABLE
from g2p.mappings.langs import __file__ as LANGS_FILE
from g2p.mappings.utils import (
    MAPPING_TYPE,
    NORM_FORM_ENUM,
    RULE_ORDERING_ENUM,
    CompactJSONMappingEncoder,
    IndentDumper,
    Rule,
    _MappingModelDefinition,
    create_fixed_width_lookbehind,
    escape_special_characters,
    expand_abbreviations,
    has_neural_support,
    load_abbreviations_from_file,
    load_alignments_from_file,
    load_from_file,
    normalize,
    strip_index_notation,
)

GEN_DIR = os.path.join(os.path.dirname(LANGS_FILE), "generated")


class Mapping(_MappingModelDefinition):
    """Class for lookup tables"""

    def model_post_init(self, *_args, **_kwargs) -> None:
        """After the model is constructed, we process the model specs by
        applying all the configuration to the rules (ie prevent feeding,
        unicode normalization etc..)"""

        if self.type == MAPPING_TYPE.mapping or self.type is None:
            # load abbreviations from path
            if self.abbreviations_path is not None and not self.abbreviations:
                self.abbreviations = load_abbreviations_from_file(
                    self.abbreviations_path
                )
            # load rules from path
            if self.rules_path is not None and not self.rules:
                # make sure self.rules is always a List[Rule] like we say it is!
                self.rules = [Rule(**obj) for obj in load_from_file(self.rules_path)]
            # Process the rules, keeping only non-empty ones, and
            # expanding abbreviations.  This is also required so that
            # we don't keep escaping special characters for example
            self.rules = self.process_model_specs()
        elif self.type == MAPPING_TYPE.lexicon:
            # load alignments from path
            if self.alignments_path is not None and not self.alignments:
                self.alignments = load_alignments_from_file(self.alignments_path)
        else:
            self.rules = []

    @staticmethod
    def find_mapping(
        in_lang: Union[None, str] = None,
        out_lang: Union[None, str] = None,
        neural: bool = False,
    ) -> "Mapping":
        """Given an input and an output language, find a mapping to get between them."""
        # Early checking for if neural is available:
        if neural and not has_neural_support():
            raise NeuralDependencyError()

        if in_lang is None or out_lang is None:
            raise exceptions.MappingMissing(in_lang, out_lang)
        fallback_mapping = None
        for mapping in MAPPINGS_AVAILABLE:
            # Don't return a neural mapping if neural is False
            if not neural and mapping.type == MAPPING_TYPE.neural:
                continue
            if mapping.in_lang == in_lang and mapping.out_lang == out_lang:
                if neural and mapping.type == MAPPING_TYPE.neural:
                    return deepcopy(mapping)
                if mapping.type == MAPPING_TYPE.lexicon:
                    # do *not* deep copy this, because alignments are big!
                    fallback_mapping = mapping.model_copy()
                else:
                    fallback_mapping = deepcopy(mapping)

        if fallback_mapping is None:
            raise exceptions.MappingMissing(in_lang, out_lang)
        else:
            return fallback_mapping

    @staticmethod
    def find_mapping_by_id(map_id: str) -> "Mapping":
        """Find the mapping with a given ID, i.e., the "id" found in the
        mapping, like in the "panphon_preprocessor" mapping."""
        for mapping in MAPPINGS_AVAILABLE:
            if mapping.id == map_id:
                return deepcopy(mapping)
        raise exceptions.MappingMissing(map_id, None)

    @staticmethod
    def load_mapping_from_path(path_to_mapping_config: Union[str, Path], index=0):
        """Loads a mapping from a path, if there is more than one mapping,
        then it loads based on the int provided to the 'index'
        argument. Default is 0."""
        mapping_config = MappingConfig.load_mapping_config_from_path(
            path_to_mapping_config
        )
        return mapping_config.mappings[index]

    @staticmethod
    def _string_to_pua(string: str, offset: int) -> str:
        """Given an string of length n, and an offset m,
           produce a string of n * chr(983040 + m).
           This makes use of the Supplementary Private Use Area A Unicode block.

        Args:
            string (str): The string to convert
            offset (int): The offset from the start of the Supplementary Private Use Area

        Returns:
            str: The resulting string
        """
        intermediate_char = chr(983040 + offset)
        prev_end = 0
        result = ""
        for match in re.finditer(r"{\d+}|$", string):
            result += intermediate_char * (match.start() - prev_end) + match.group()
            prev_end = match.end()
        return result

    def index(self, item):
        """Find the location of an item in self"""
        return self.rules.index(item)

    def inventory(self, in_or_out: str = "in", non_empty: bool = False):
        """Return just inputs or outputs as inventory of mapping"""
        if in_or_out == "in":
            in_or_out = "rule_input"
        if in_or_out == "out":
            in_or_out = "rule_output"
        inv = [getattr(x, in_or_out) for x in self.rules]
        if non_empty:
            return [sym for sym in inv if sym != ""]
        else:
            return inv

    def plain_mapping(self):
        """Return the plain mapping for displaying or saving to disk.

        Args:
            skip_empty_contexts: when set, filter out empty context_before/after
        """
        return [rule.export_to_dict() for rule in self.rules]

    def process_model_specs(self) -> List[Rule]:
        """Process all model specifications"""
        if self.as_is is not None:
            appropriate_setting = (
                RULE_ORDERING_ENUM.as_written
                if self.as_is
                else RULE_ORDERING_ENUM.apply_longest_first
            )
            self.rule_ordering = appropriate_setting

            LOGGER.warning(
                f"mapping from {self.in_lang} to {self.out_lang} "
                'is using the deprecated parameter "as_is"; '
                f"replace `as_is: {self.as_is}` with `rule_ordering: {appropriate_setting.value}`"
            )
            if version_tuple < (3,):
                LOGGER.warning(
                    "as_is support will be removed in the next major version, g2p 3."
                )
            else:
                LOGGER.error("The as_is feature has been removed in version 3.0.0")
                import sys

                sys.exit(1)

        # Sorting must happen before the calculation of PUA intermediate forms for proper indexing
        if self.rule_ordering == RULE_ORDERING_ENUM.apply_longest_first:
            self.rules = sorted(
                # Temporarily normalize to NFD for heuristic sorting of NFC-defined rules
                self.rules,
                key=lambda x: len(normalize(strip_index_notation(x.rule_input), "NFD")),
                reverse=True,
            )

        def apply_to_attributes(rule: Rule, func: Callable, *attrs):
            for k in attrs:
                value = getattr(rule, k)
                if value:  # won't be None since default is ""
                    setattr(rule, k, func(value))

        non_empty_mappings: List[Rule] = []
        for i, rule in enumerate(self.rules):
            # We explicitly exclude match_pattern and
            # intermediate_form when saving rules.  Seeing either of
            # them is a programmer error.
            assert (
                rule.match_pattern is None
            ), "Either match_pattern was specified explicitly or process_model_specs was called more than once"
            assert (
                rule.intermediate_form is None
            ), "Either intermediate_form was specified explicitly or process_model_specs was called more than once"
            # Expand Abbreviations
            if self.abbreviations:
                apply_to_attributes(
                    rule,
                    partial(expand_abbreviations, abbs=self.abbreviations),
                    "rule_input",
                    "context_before",
                    "context_after",
                )
            # Reverse Rule
            if self.reverse:
                rule.rule_input, rule.rule_output = rule.rule_output, rule.rule_input
                rule.context_before = ""
                rule.context_after = ""
            # Escape Special
            if not self.processed and self.escape_special:
                rule = escape_special_characters(rule)
            # Unicode Normalization
            if self.norm_form != NORM_FORM_ENUM.none:
                apply_to_attributes(
                    rule,
                    partial(normalize, norm_form=self.norm_form.value),
                    "rule_input",
                    "rule_output",
                    "context_before",
                    "context_after",
                )
            # Prevent Feeding
            if self.prevent_feeding or rule.prevent_feeding:
                rule.intermediate_form = self._string_to_pua(rule.rule_output, i)
            # Create match pattern
            rule.match_pattern = self.rule_to_regex(rule)
            # Only add non-empty rules
            if rule.match_pattern:
                non_empty_mappings.append(rule)

        self.processed = True
        return non_empty_mappings

    def rule_to_regex(self, rule: Union[Rule, dict]) -> Union[Pattern, None]:
        """Turns an input string (and the context) from an input/output pair
        into a regular expression pattern"

        The 'in' key is the match.
        The 'context_after' key creates a lookahead.
        The 'context_before' key creates a lookbehind.

        Args:
            rule: A dictionary containing 'in', 'out', 'context_before', and 'context_after' keys

        Raises:
            Exception: This is raised when un-supported regex characters or symbols exist in the rule

        Returns:
            Pattern: returns a regex pattern (re.Pattern)
            None: if input is null
        """
        # Prevent null input. See, https://github.com/NRC-ILT/g2p/issues/24
        if isinstance(rule, dict):
            rule = Rule(**rule)
        if not rule.rule_input:
            LOGGER.warning(
                f"Rule {rule} has no input. This is disallowed. "
                "Please check your mapping file for rules with null inputs."
            )
            return None
        input_match = strip_index_notation(rule.rule_input)
        try:
            inp = create_fixed_width_lookbehind(rule.context_before) + input_match
            if rule.context_after:
                inp += f"(?={rule.context_after})"
            if not self.case_sensitive:
                rule_regex = re.compile(inp, re.I)
            else:
                rule_regex = re.compile(inp)
        except re.error as e:
            in_lang = self.in_lang
            out_lang = self.out_lang
            LOGGER.error(
                "Your regex in mapping between %s and %s is malformed.  "
                "Do you have un-escaped regex characters in your input %s, contexts %s, %s?  "
                "Error is: %s",
                in_lang,
                out_lang,
                inp,
                rule.context_before,
                rule.context_after,
                e.msg,
            )
            raise exceptions.MalformedMapping(
                f"Your regex in mapping between {in_lang} and {out_lang} is malformed.  "
                f"Do you have un-escaped regex characters in your input {inp}, "
                f"contexts {rule.context_before}, {rule.context_after}?"
            ) from e
        return rule_regex

    def extend(self, mapping: "Mapping"):
        """Add all the rules from mapping into self, effectively merging two mappings

        Caveat: if self and mapping have contradictory rules, which one will
        "win" is unspecified, and may depend on mapping configuration options.
        """
        self.rules.extend(mapping.rules)

    def deduplicate(self):
        """Remove duplicate rules found in self, keeping the first copy found."""
        # Since Python 3.6, dict keeps its element in insertion order (while
        # set does not), so deduplicating the rules is a one-liner:
        self.rules = list({repr(rule): rule for rule in self.rules}.values())

    def mapping_to_stream(self, out_stream, file_type: str = "json"):
        """Write mapping to a stream"""

        if file_type == "json":
            json.dump(
                self.plain_mapping(),
                out_stream,
                indent=4,
                ensure_ascii=False,
                cls=CompactJSONMappingEncoder,
            )
            print("\n", end="", file=out_stream)
        elif file_type == "csv":
            fieldnames = ["in", "out", "context_before", "context_after"]
            writer = csv.DictWriter(
                out_stream, fieldnames=fieldnames, extrasaction="ignore"
            )
            for io in self.rules:
                writer.writerow(io.export_to_dict())
        else:
            raise exceptions.IncorrectFileType(f"File type {file_type} is invalid.")

    def mapping_to_file(self, output_path: str = GEN_DIR, file_type: str = "json"):
        """Write mapping to file"""

        if not os.path.isdir(output_path):
            raise Exception(f"Path {output_path} is not a directory")
        fn = os.path.join(
            output_path,
            self.in_lang + "_to_" + self.out_lang + "." + file_type,
        )
        with open(fn, "w", encoding="utf8", newline="\n") as f:
            self.mapping_to_stream(f, file_type)

    def export_to_dict(self, mapping_type="json", config_only=False):
        """Export a mapping to a dictionary, optionally including only the
        configuration (and not the rules or alignments) in the case
        where we are just writing the config file.
        """
        if config_only:
            model_dict = self.model_dump(
                mode="json",
                exclude_none=True,
                exclude={
                    "parent_dir": True,
                    "rules": True,
                    "processed": True,
                    "alignments": True,
                    "abbreviations": True,
                },
            )
        else:
            model_dict = self.model_dump(
                mode="json",
                exclude_none=True,
                exclude={"parent_dir": True},
            )
        if not model_dict.get("rules_path"):
            model_dict["rules_path"] = (
                f"{self.in_lang}_to_{self.out_lang}.{mapping_type}"
            )
        return model_dict

    def config_to_file(
        self,
        output_path: str = os.path.join(GEN_DIR, "config-g2p.yaml"),
    ):
        """Write configuration to file."""
        add_config = False
        if os.path.isdir(output_path):
            output_path = os.path.join(output_path, "config-g2p.yaml")
        if os.path.exists(output_path) and os.path.isfile(output_path):
            LOGGER.warning(f"Adding mapping config to file at {output_path}")
            add_config = True
        else:
            LOGGER.warning(f"writing mapping config to file at {output_path}")
        fn = output_path
        config_template = self.export_to_dict(config_only=True)
        # Serialize piece-by-piece, which is why this is a list of type dict and not type Mapping
        # If config file exists already, just add the mapping.
        to_export = None
        if add_config:
            existing_data = MappingConfig.load_mapping_config_from_path(fn)
            updated = False
            for i, mapping in enumerate(existing_data.mappings):
                # if the mapping exists, just update the generation data
                if (
                    mapping.in_lang == config_template["in_lang"]
                    and mapping.out_lang == config_template["out_lang"]
                ):
                    existing_data.mappings[i].authors = config_template["authors"]
                    updated = True
                    break
            if not updated:
                existing_data.mappings.append(config_template)
            to_export = {
                "mappings": [
                    x.export_to_dict(config_only=True) if isinstance(x, Mapping) else x
                    for x in existing_data.mappings
                ],
            }
        else:
            to_export = {"mappings": [config_template]}
        with open(fn, "w", encoding="utf8", newline="\n") as f:
            yaml.dump(
                to_export,
                f,
                Dumper=IndentDumper,
                default_flow_style=False,
                # do not write strings as unreadable \u escapes! (see
                # https://stackoverflow.com/questions/10648614/dump-in-pyyaml-as-utf-8)
                allow_unicode=True,
            )


MAPPINGS_AVAILABLE: List[Mapping] = [
    Mapping(**mapping) for mapping in _MAPPINGS_AVAILABLE
]


class MappingConfig(BaseModel):
    """This is the format used by g2p for configuring mappings."""

    mappings: List[Mapping]

    def export_to_dict(self):
        return {"mappings": [mapping.export_to_dict() for mapping in self.mappings]}

    @staticmethod
    def load_mapping_config_from_path(
        path_to_mapping_config: Union[str, Path],
    ) -> "MappingConfig":
        """Loads a mapping configuration from a path, if you just want one specific mapping
        from the config, you can try Mapping.load_mapping_from_path instead.
        """
        if isinstance(path_to_mapping_config, str):
            path = Path(path_to_mapping_config)
        else:
            path = path_to_mapping_config
        parent_dir = path.parent
        with open(path, encoding="utf8") as f:
            loaded_config = yaml.safe_load(f)
            if "mappings" in loaded_config:
                for mapping in loaded_config["mappings"]:
                    mapping["parent_dir"] = parent_dir
        try:
            return MappingConfig(**loaded_config)
        except exceptions.MalformedMapping as e:
            e.message = f"{e.message}\nProblem in config file: {path}"
            raise e
        except TypeError as e:
            raise exceptions.MalformedMapping("Config file: {path}") from e


LANGS: Dict[str, MappingConfig] = {k: MappingConfig(**v) for k, v in _LANGS.items()}



######################################################################
# Patrick Littell
#
# create_ipa_mapping.py
#
# Given two IPA inventories in JSON (either as dedicated inventory
# files or the input/output sides of mapping files), map the first
# onto the second by use of panphon's phonetic distance calculators.
#
# The resulting mappings are used just like other mappings: to make
# converters and pipelines of converters in convert_orthography.py
#
# AP Note: Taken from ReadAlongs-Studio and implemented with G2P formatting
######################################################################
import datetime as dt
from typing import Iterable, List, Tuple

from tqdm import tqdm

from g2p.constants import DISTANCE_METRICS
from g2p.log import LOGGER
from g2p.mappings import Mapping
from g2p.mappings.langs.utils import getPanphonDistanceSingleton
from g2p.mappings.utils import RULE_ORDERING_ENUM, is_ipa, is_xsampa
from g2p.transducer import Transducer

#################################
#
# Preprocessing:
#
# Panphon can only match a single segment to another segment, rather
# than (say) try to combine two segments to better match the features.
# For example, you might want "kʷ" to match to "kw", but Panphon will
# only match the "kʷ" to "k" and consider the "w" to be a dropped
# character.  In order to get around this, we preprocess strings so
# that common IPA segments that you might expect map to two characters
# in another language, like affricates or rounded consonants, are
# treated as two characters rather than one.
#
#################################
_xsampa_converter = None  # Cache this but create it only if needed


def process_character(p, is_xsampa=False):
    if is_xsampa:
        global _xsampa_converter
        if _xsampa_converter is None:
            # Expensive import, do it only when needed:
            from panphon.xsampa import XSampa  # type: ignore

            _xsampa_converter = XSampa()
        p = _xsampa_converter.convert(p)
    panphon_preprocessor = Transducer(
        Mapping.find_mapping_by_id("panphon_preprocessor")
    )
    return panphon_preprocessor(p).output_string


def process_characters(inv, is_xsampa=False):
    return [process_character(p, is_xsampa) for p in inv]


##################################
#
# Creating the mapping
#
#
#
###################################


def get_distance_method(dst, distance: str):
    if distance not in DISTANCE_METRICS:
        raise ValueError(f"Distance metric {distance} not supported")
    try:
        distance_method = getattr(dst, distance)
    except AttributeError as e:
        # Older versions of panphon mispelled Dolgopolsky's name as Dogolpolsky...
        # Try again with the older name, so we stay compatible with both <=0.19
        # and >=0.19.1
        if distance == "dolgo_prime_distance":
            return getattr(dst, "dogol_prime_distance")

        LOGGER.error(f"The distance metric {distance} is not supported by PanPhon")
        raise ValueError(f"Distance metric {distance} not supported") from e
    return distance_method


def create_multi_mapping(
    src_mappings: List[Tuple[Mapping, str]],
    tgt_mappings: List[Tuple[Mapping, str]],
    distance: str = "weighted_feature_edit_distance",
    quiet=False,
) -> Mapping:
    """Create a mapping for a set of source mappings to a set of target mappings

    Each src/tgt mappings is a (mapping: Mapping, in_or_out: str) pair specifying
    the mapping to use and whether its input ("in") or output ("out") inventory
    should be used to create the new mapping.

    The name of the mapping is infered from src_mappings[0] and tgt_mappings[0]'s
    metadata.
    """

    def compact_ipa_names(ipa_names: Iterable) -> str:
        # ["fra-ipa", "eng-ipa", "kwk-ipa"] -> "fra-eng-kwk-ipa"
        return (
            "-".join(name[:-4] if name.endswith("-ipa") else name for name in ipa_names)
            + "-ipa"
        )

    def long_ipa_names(ipa_names: Iterable) -> str:
        # ["fra-ipa", "eng-ipa", "kwk-ipa"] -> "fra-ipa and eng-ipa and kwk-ipa"
        return " and ".join(ipa_names)

    def get_sorted_unique_names(mappings: List[Tuple[Mapping, str]]) -> List[str]:
        return sorted(
            {getattr(mapping, f"{in_or_out}_lang") for mapping, in_or_out in mappings}
        )

    def deduplicate(iterable: Iterable) -> List:
        # Use a dict, and not a set, to preserve the original order.
        return list({v: v for v in iterable}.values())

    map_1_names = get_sorted_unique_names(src_mappings)
    map_2_names = get_sorted_unique_names(tgt_mappings)

    src_inventory = []
    for mapping, io in src_mappings:
        name = getattr(mapping, f"{io}_lang")
        if not is_ipa(name):
            LOGGER.warning(
                "Unsupported orthography of src inventory: %s; must be IPA", name
            )
        src_inventory.extend(mapping.inventory(io))
    src_inventory = deduplicate(src_inventory)

    tgt_inventory = []
    for mapping, io in tgt_mappings:
        name = getattr(mapping, f"{io}_lang")
        if not is_ipa(name):
            LOGGER.warning(
                "Unsupported orthography of tgt inventory: %s; must be IPA", name
            )
        tgt_inventory.extend(mapping.inventory(io))
    tgt_inventory = deduplicate(tgt_inventory)

    mapping = align_inventories(
        src_inventory, tgt_inventory, distance=distance, quiet=quiet
    )

    config = {
        "in_lang": compact_ipa_names(map_1_names),
        "out_lang": compact_ipa_names(map_2_names),
        "language_name": "IPA",
        "rule_ordering": RULE_ORDERING_ENUM.apply_longest_first.value,
        "rules": mapping,
        "prevent_feeding": True,
        "norm_form": "NFC",
        "authors": [f"Generated {dt.datetime.now()}"],
        "display_name": (
            long_ipa_names(map_1_names) + " to " + long_ipa_names(map_2_names)
        ),
    }

    return Mapping(**config)


def create_mapping(
    mapping_1: Mapping,
    mapping_2: Mapping,
    mapping_1_io: str = "out",
    mapping_2_io: str = "in",
    distance: str = "weighted_feature_edit_distance",
    quiet=False,
) -> Mapping:
    """Create a mapping from mapping_1's output inventory to mapping_2's input inventory"""

    map_1_name = getattr(mapping_1, f"{mapping_1_io}_lang")
    map_2_name = getattr(mapping_2, f"{mapping_2_io}_lang")
    if not is_ipa(map_1_name) and not is_xsampa(map_1_name):
        LOGGER.warning(
            "Unsupported orthography of inventory 1: %s (must be ipa or x-sampa)",
            map_1_name,
        )
    if not is_ipa(map_2_name) and not is_xsampa(map_2_name):
        LOGGER.warning(
            "Unsupported orthography of inventory 2: %s (must be ipa or x-sampa)",
            map_2_name,
        )
    l1_is_xsampa, l2_is_xsampa = is_xsampa(map_1_name), is_xsampa(map_2_name)
    rules = align_inventories(
        mapping_1.inventory(mapping_1_io, non_empty=True),
        mapping_2.inventory(mapping_2_io, non_empty=True),
        l1_is_xsampa,
        l2_is_xsampa,
        distance=distance,
        quiet=quiet,
    )

    # Initialize mapping with input language parameters (as_is,
    # case_sensitive, prevent_feeding, etc) but *not* the same filename!
    config = mapping_1.model_copy().model_dump(exclude={"rules_path"})
    config = {
        **config,
        "authors": [f"Generated {dt.datetime.now()}"],
        "display_name": None,
        "language_name": None,
        "in_lang": map_1_name,
        "out_lang": map_2_name,
        "rules": rules,
        "prevent_feeding": True,
        "rule_ordering": RULE_ORDERING_ENUM.apply_longest_first,
    }

    return Mapping(**config)


def align_inventories(
    inventory_l1,
    inventory_l2,
    l1_is_xsampa=False,
    l2_is_xsampa=False,
    distance="weighted_feature_edit_distance",
    quiet=False,
):
    """Align inventories by finding a good sequence in inventory_l2 for each
    character in inventory_l1"""

    # find_good_match() is a function inside align_inventories() because it
    # lets us initialize dst and ps_pseqs globally once, yielding a roughly 8x
    # speed inprovements over the previous version of the code.

    # Initializing panphon.distance.Distance() is expensive, so do it just once
    dst = getPanphonDistanceSingleton()
    # Initializing p2_pseqs is expensive, so do it only once per call to align_inventories()
    p2_pseqs = [
        dst.fm.ipa_segs(p) for p in process_characters(inventory_l2, l2_is_xsampa)
    ]

    def find_good_match(p1, inventory_l2):
        """Find a good sequence in inventory_l2 matching p1."""

        # The proper way to do this would be with some kind of beam search
        # through a determinized/minimized FST, but in the absence of that
        # we can do a kind of heurstic greedy search.  (we don't want any
        # dependencies outside of PyPI otherwise we'd just use OpenFST)

        p1_pseq = dst.fm.ipa_segs(p1)

        i = 0
        good_match = []
        while i < len(p1_pseq):
            best_input = ""
            best_output = -1
            best_score = 0xDEADBEEF
            for j, p2_pseq in enumerate(p2_pseqs):
                # FIXME: Should also consider the (weighted) possibility
                # of deleting input or inserting any segment (but that
                # can't be done with a greedy search)
                if len(p2_pseq) == 0:
                    LOGGER.warning(
                        "No panphon mapping for %s - skipping", inventory_l2[j]
                    )
                    continue
                e = min(i + len(p2_pseq), len(p1_pseq))
                input_seg = p1_pseq[i:e]
                distance_method = get_distance_method(dst, distance)
                score = distance_method("".join(input_seg), "".join(p2_pseq))
                # Be very greedy and take the longest match
                if (
                    score < best_score
                    or score == best_score
                    and len(input_seg) > len(best_input)
                ):
                    best_input = input_seg
                    best_output = j
                    best_score = score
            LOGGER.debug(
                "Best match at position %d: %s => %s",
                i,
                best_input,
                inventory_l2[best_output],
            )
            good_match.append(inventory_l2[best_output])
            i += len(best_input)  # greedy!
        return "".join(good_match)

    mapping = []
    if not quiet:
        pbar = tqdm(total=100)
    step = 1 / len(inventory_l1) * 100
    for i1, p1 in enumerate(process_characters(inventory_l1, l1_is_xsampa)):
        # we enumerate the strings because we want to save the original string
        # (e.g., 'kʷ') to the mapping, not the processed one (e.g. 'kw')
        good_match = find_good_match(p1, inventory_l2)
        mapping.append({"in": inventory_l1[i1], "out": good_match})
        if not quiet:
            pbar.update(step)
    if not quiet:
        pbar.close()
    return mapping



#!/usr/bin/env python

from unittest import TestCase, main

from g2p import make_g2p
from g2p.mappings import Mapping
from g2p.mappings.utils import normalize
from g2p.transducer import Transducer


class UnidecodeTransducerTest(TestCase):
    def test_unidecode_mapping(self):
        m = Mapping(type="unidecode")
        self.assertEqual(m.rules, [])
        self.assertEqual(m.type, "unidecode")
        t = Transducer(m)
        tg = t("été Nunavut ᓄᓇᕗᑦ")
        self.assertEqual(tg.output_string, "ete Nunavut nonafot")

    def test_unidecode_g2p(self):
        transducer = make_g2p("und", "und-ascii", tokenize=False)
        tg = transducer(normalize("éçà", "NFD"))
        self.assertEqual(tg.output_string, "eca")
        self.assertEqual(tg.edges, [(0, 0), (1, 0), (2, 1), (3, 1), (4, 2), (5, 2)])

        tg = transducer(normalize("éçà", "NFC"))
        self.assertEqual(tg.output_string, "eca")
        self.assertEqual(tg.edges, [(0, 0), (1, 1), (2, 2)])

    def test_unidecode_empty_output(self):
        transducer = make_g2p("und", "und-ascii", tokenize=False)
        # \u0361 on its own gets deleted completely by unidecode
        tg = transducer("\u0361")
        self.assertEqual(tg.output_string, "")
        self.assertEqual(tg.edges, [])

    def test_unidecode_to_arpabet(self):
        transducer = make_g2p("und", "eng-arpabet", tokenize=False)
        tg = transducer("été Nunavut ᓄᓇᕗᑦ")
        self.assertEqual(
            tg.output_string, "EY T EY  N UW N AA V UW T  N OW N AA F OW T "
        )

    def test_unidecode_arabic_to_arpabet(self):
        transducer = make_g2p("und", "eng-arpabet")
        tg = transducer("السلام عليكم")
        self.assertEqual(tg.output_string, "L S L M  L Y K M ")

    def test_unidecode_arabic_presentation_to_arpabet(self):
        transducer = make_g2p("und", "eng-arpabet")
        tg = transducer("ﺷﻜﺮﺍﹰ")
        self.assertEqual(tg.output_string, "S HH K D AA N ")

    def test_unidecode_kanji_to_arpabet(self):
        transducer = make_g2p("und", "eng-arpabet")
        tg = transducer("日本語")
        self.assertEqual(tg.output_string, "D IY B EY N Y UW ")

    def test_unidecode_hanzi_to_arpabet(self):
        transducer = make_g2p("und", "eng-arpabet", tokenize=False)
        tg = transducer("你们好!你们说汉语马?")
        self.assertEqual(
            tg.output_string,
            "N IY M EY N HH AA OW N IY M EY N S HH UW OW Y IY Y UW M AA HH ",
        )


if __name__ == "__main__":
    main()



"""REST API for G2P index-preserving grapheme-to-phoneme conversion using FastAPI."""

from enum import Enum
from typing import List

from fastapi import FastAPI, HTTPException, Path, Query
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse, PlainTextResponse

from g2p import make_g2p
from g2p.exceptions import NoPath
from g2p.log import LOGGER
from g2p.mappings.langs import LANGS_NETWORK

# Create the v1 version of the API
api = FastAPI(
    title="Simple G2P API",
    description="A simple API for the G2P module",
    version="1.1.1",
    contact={"email": "readalong.studio@gmail.com"},
    license_info={
        "name": "MIT",
        "url": "https://github.com/NRC-ILT/g2p/blob/main/LICENSE",
    },
    openapi_tags=[
        {
            "name": "ancestors",
            "description": "Find which mappings can convert to a given node",
        },
        {
            "name": "descendants",
            "description": "Find which mappings can be converted to from a given node",
        },
        {"name": "g2p", "description": "Transduced, g2p'ed forms"},
        {"name": "langs", "description": "Languages/mappings available for G2P"},
    ],
)

# Get the langs
LANGS = sorted(LANGS_NETWORK.nodes)
Lang = Enum("Lang", [(name, name) for name in LANGS])  # type: ignore


# Be compatible with previous API which returned 404 on an unknown node
@api.exception_handler(RequestValidationError)
async def validation_exception_handler(request, exc: RequestValidationError):
    error = exc.errors()[0]
    LOGGER.error("%s", error.get("msg", "Unknown Error"))
    if error.get("type") == "enum":
        return PlainTextResponse(
            "Unknown input or output language code", status_code=404
        )
    else:
        return JSONResponse(
            status_code=422,
            content={"detail": exc.errors(), "body": exc.body},
        )


@api.get(
    "/ancestors/{node}",
    summary="get all ancestors of node",
    tags=["ancestors"],
    operation_id="getAncestors",
    response_description="The valid ancestors of a node",
)
def get_all_ancestors_of_node(
    node: Lang = Path(description="language node name"),
) -> List[str]:
    """Get the valid ancestors in the network's path to a given node. These
    are all the mappings that you can convert from in order to get the
    given node."""
    return sorted(LANGS_NETWORK.ancestors(node.name))


@api.get(
    "/descendants/{node}",
    summary="get all descendants of node",
    tags=["descendants"],
    operation_id="getDescendants",
    response_description="The valid descendants of a node",
)
def get_all_descendants_of_node(
    node: Lang = Path(description="language node name"),
) -> List[str]:
    return sorted(LANGS_NETWORK.descendants(node.name))


@api.get(
    "/g2p",
    summary="get g2p'ed form",
    tags=["g2p"],
    operation_id="convertString",
    response_description="The converted text",
)
def g2p(
    in_lang: Lang = Query(alias="in-lang", description="input lang of string"),
    out_lang: Lang = Query(alias="out-lang", description="output lang of string"),
    text: str = Query(description="string to convert"),
    index: bool = Query(False, description="return indices"),
    debugger: bool = Query(False, description="return debugging information"),
    tokenize: bool = Query(
        False,
        description="tokenize before transducing; warning: set this to true unless your text is already a single word without punctuation",
    ),
) -> dict:
    """Get the converted version of a string, given an input and output lang"""
    try:
        transducer = make_g2p(in_lang.name, out_lang.name, tokenize=tokenize)
        tg = transducer(text)
        return {
            "input-text": tg.input_string,
            "output-text": tg.output_string,
            "debugger": tg.debugger if debugger else debugger,
            "index": tg.edges if index else index,
        }
    except NoPath:
        raise HTTPException(
            status_code=400, detail=f"No path from {in_lang} to {out_lang}"
        )


@api.get(
    "/langs",
    summary="find all possible languages in g2p",
    tags=["langs"],
    operation_id="searchTable",
    response_description="search results matching criteria",
)
def langs() -> List[str]:
    """By passing in the appropriate options, you can find available mappings"""
    return LANGS



"""REST API for G2P index-preserving grapheme-to-phoneme conversion using FastAPI.

You can run the API app for development purposes on any platform with:
    pip install uvicorn fastapi
    DEVELOPMENT=1 uvicorn g2p.api_v2:api --reload --port 5000
- The --reload switch will watch for changes under the directory where it's
  running and reload the code whenever it changes.
- DEVELOPMENT=1 tells the API to accept cross-origin requests (i.e. by sending the
  appropriate CORS headers) from development servers running on localhost, e.g.
  http://localhost:4200

For deployment, you can use the ORIGIN environment variable to set the
URL of your application in order to make it accept requests from that
site.  For instance if you deployed an application that uses it at
https://my.awesome.site you would set ORIGIN=https://my.awesome.site
in your environment variables.  This is usually done through an
environment variable file (or in a dashboard) and will depend on your
hosting environment.

You can also spin up the API server grade (on Linux, not Windows) with gunicorn:
    pip install -r requirements.api.txt
    gunicorn -b 127.0.0.1:5000 -w 4 -k uvicorn.workers.UvicornWorker g2p.api_v2:api

Once spun up, the API will be visible at
http://localhost:5000/docs

"""

import os
from enum import Enum
from typing import Dict, List, Tuple, Union

from fastapi import Body, FastAPI, HTTPException, Path
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel, Field

import g2p
import g2p.mappings as g2p_mappings
import g2p.mappings.langs as g2p_langs
from g2p.exceptions import InvalidLanguageCode, NoPath
from g2p.log import LOGGER
from g2p.transducer import CompositeTransductionGraph, TransductionGraph

# Create the v2 version of the API
api = FastAPI(
    title="Simple G2P API",
    description="A simple API for the G2P module",
    version="2.0.0",
    contact={"email": "readalong.studio@gmail.com"},
    license_info={
        "name": "MIT",
        "url": "https://github.com/NRC-ILT/g2p/blob/main/LICENSE",
    },
)
# Allow for development/non-g2p-studio deployments
middleware_args: Dict[str, Union[str, List[str]]] = {}
if os.getenv("DEVELOPMENT", False):  # pragma: no cover
    LOGGER.info(
        "Running in development mode, will allow requests from http://localhost:*"
    )
    # Allow requests from localhost dev servers
    middleware_args["allow_origin_regex"] = "http://localhost(:.*)?"
else:
    # Allow requests from a configured site name
    origin = os.getenv("ORIGIN")
    if origin is not None:
        middleware_args["allow_origins"] = origin.split()
        LOGGER.info("Allowing requests from the following sites:")
        for site in middleware_args["allow_origins"]:
            LOGGER.info("   %s", site)
# If DEVELOPMENT or ORIGIN are not specified then same-origin only
if middleware_args:
    api.add_middleware(
        CORSMiddleware, allow_methods=["GET", "POST", "OPTIONS"], **middleware_args
    )

# All possible language codes
LanguageNode = Enum("LanguageNode", [(name, name) for name in g2p_langs.LANGS_NETWORK.nodes])  # type: ignore


class SupportedLanguage(BaseModel):
    """Writing or phonetic system for conversion"""

    code: str = Field(
        description="Language, writing, or phonetic system code as passed to /convert",
        examples=["eng"],
    )
    name: Union[str, None] = Field(
        description="Display name (not internationalized) for language, may be None",
        examples=["English"],
    )


MAPPING_NAMES: Dict[str, Union[str, None]] = {}
for mapping in g2p_mappings.MAPPINGS_AVAILABLE:
    MAPPING_NAMES[mapping.in_lang] = mapping.language_name
LANGS = []
for code, config in g2p_mappings.LANGS.items():
    if code not in ("generated", "font-encodings"):
        language_name = config.mappings[0].language_name
        if MAPPING_NAMES.get(code) is None:
            MAPPING_NAMES[code] = language_name
        LANGS.append(SupportedLanguage(code=code, name=language_name))
LANGS.sort(key=lambda x: str(x.name).lower())
CODES = []
for code in g2p_langs.LANGS_NETWORK.nodes:
    CODES.append(SupportedLanguage(code=code, name=MAPPING_NAMES.get(code, None)))
CODES.sort(key=lambda x: x.code)


@api.get(
    "/langs",
    response_description="Supported languages for conversion",
)
def get_supported_input_languages() -> List[SupportedLanguage]:
    """Return a list of language codes (possible inputs for /convert) and their names."""
    return LANGS


@api.get(
    "/nodes",
    response_description="Supported writing or phonetic systems for conversion",
)
def get_all_writing_or_phonetic_systems() -> List[SupportedLanguage]:
    """Return a list of all possible inputs for /convert and
    associated language names, if any.
    """
    return CODES


class ConvertRequest(BaseModel):
    """Request conversion from one writing or phonetic system to another."""

    in_lang: LanguageNode = Field(
        description="Name of input node", examples=["eng-ipa"]
    )
    out_lang: LanguageNode = Field(
        description="Name of output node", examples=["eng-arpabet"]
    )
    text: str = Field(description="Text to convert", examples=["hɛloʊ"])
    tokenize: Union[bool, None] = Field(
        True,
        description="Tokenize input and return a list of segments.  This is "
        "the default behaviour, set this to `False` to treat input as a single segment. "
        "Warning: only use False if text is already a single word without punctuation.",
    )
    compose_from: Union[str, None] = Field(
        None,
        description="Compose all conversions from a specific step onwards."
        "  Useful if you wish to recover alignments for a specific intermediate step",
    )
    indices: Union[bool, None] = Field(
        False,
        description="Return lists of input and output characters and alignments "
        "as indices into those lists.  These are guaranteed to be usable no matter "
        "how your platform defines a 'character'.",
    )


class Conversion(BaseModel):
    """One step in G2P conversion"""

    in_lang: Union[None, LanguageNode] = Field(
        None,
        description="Name of input language, absent if no conversion was done",
        examples=["eng-ipa"],
    )
    out_lang: Union[None, LanguageNode] = Field(
        None,
        description="Name of output language, absent if no conversion was done",
        examples=["eng-arpabet"],
    )
    input_nodes: Union[None, List[str]] = Field(
        None,
        description="Characters in input, which can be safely indexed by alignments, "
        "present only if `indices` was True in request.",
        examples=[["h", "i", "🙂", "🙂", "🙂"]],
    )
    output_nodes: Union[None, List[str]] = Field(
        None,
        description="Characters in output, which can be safely indexed by alignments, "
        "present only if `indices` was True in request.",
        examples=[["H", "H", " ", "I", "Y", " ", "🙂", "🙂", "🙂"]],
    )
    alignments: Union[None, List[Tuple[int, Union[int, None]]]] = Field(
        None,
        description="Alignments from input to output indices, "
        "present only if `indices` was True in request.",
        examples=[
            [
                [0, 0],
                [0, 1],
                [0, 2],
                [1, 3],
                [1, 4],
                [1, 5],
                [2, 6],
                [2, 7],
                [3, 8],
                [3, 9],
                [3, 10],
            ]
        ],
    )
    substring_alignments: List[Tuple[str, str]] = Field(
        description="Minimal montonic substring alignments from input to output substrings",
        examples=[
            [
                ["h", "HH "],
                ["ɛ", "EH "],
                ["l", "L "],
                ["oʊ", "OW "],
            ]
        ],
    )


class Segment(BaseModel):
    """Result of G2P conversion of one segment of input, with
    intermediate steps and substring_alignments."""

    conversions: List[Conversion] = Field(
        description="Sequence of conversions in reverse order.",
        examples=[
            [
                {
                    "in_lang": "eng-ipa",
                    "out_lang": "eng-arpabet",
                    "substring_alignments": [
                        ["h", "HH "],
                        ["ɛ", "EH "],
                        ["l", "L "],
                        ["oʊ", "OW "],
                    ],
                }
            ]
        ],
    )


@api.post("/convert")
def convert_one_writing_or_phonetic_system_to_another(  # noqa: C901
    request: ConvertRequest = Body(
        openapi_examples={
            "eng-ipa to eng-arpabet": {
                "summary": "Convert English IPA to ARPABET",
                "description": "G2P can do simple conversions between equivalent phonetic notations",
                "value": {
                    "in_lang": "eng-ipa",
                    "out_lang": "eng-arpabet",
                    "text": "hɛloʊ",
                },
            },
            "fin to eng-arpabet": {
                "summary": "Convert Finnish orthography to nearest ARPABET phones",
                "description": """
G2P can do also (sometimes) approximate one language's phonology using phones from another.
To find out possible output languages for an input, use the 'outputs_for' endpoint.""",
                "value": {
                    "in_lang": "fin",
                    "out_lang": "eng-arpabet",
                    "text": "hyvää yötä",
                },
            },
            "composed conversion": {
                "summary": "Convert Finnish orthography directly to nearest ARPABET phones",
                "description": "By default all the conversion steps are returned.  "
                "You can get the direct mapping from input to output by setting compose_from: in_lang",
                "value": {
                    "in_lang": "fin",
                    "out_lang": "eng-arpabet",
                    "text": "hyvää huomenta",
                    "compose_from": "fin",
                },
            },
            "tokenized conversion": {
                "summary": "Convert Finnish orthography with punctuation directly to nearest ARPABET phones",
                "description": "Non-word segments are returned with in_lang and out_lang as null",
                "value": {
                    "in_lang": "fin",
                    "out_lang": "eng-arpabet",
                    "text": "Varaus! Sieni, joka kasvaa korvessa, on myrkyllinen!",
                    "compose_from": "fin-ipa",
                },
            },
        }
    )
) -> List[Segment]:
    """Tokenize a text return the converted and intermediate forms of each
    segment (non-token segments will have converted=False).  The final
    conversion comes first in the output, followed by prevoius
    conversions.  If you do not want the intermediate conversions, set
    "compose_from" to in_lang.

    """
    in_lang = request.in_lang.name
    out_lang = request.out_lang.name
    try:
        transducer = g2p.make_g2p(in_lang, out_lang)
        if request.tokenize:
            tokenizer = g2p.make_tokenizer(in_lang)
            tokens = tokenizer.tokenize_text(request.text)
        else:
            tokens = [g2p.Token(request.text, is_word=True)]
    except NoPath:
        raise HTTPException(
            status_code=400, detail=f"No path from {in_lang} to {out_lang}"
        )
    except InvalidLanguageCode:  # pragma: nocover
        # Will never happen due to FastAPI validation (will get 422 instead)
        raise HTTPException(
            status_code=404, detail="Unknown input or output language code"
        )

    segments: List[Segment] = []
    for token in tokens:
        conversions: List[Conversion] = []
        if not token.is_word:  # non-word, has no in_lang/out_lang
            tg = TransductionGraph(token.text)
            conv = Conversion(substring_alignments=tg.substring_alignments())
            if request.indices:
                conv.alignments = tg.alignments()
                conv.input_nodes = list(tg.input_string)
                conv.output_nodes = list(tg.output_string)
            conversions.append(conv)
        else:
            tg = transducer(token.text)
            if request.compose_from:
                composed_tiers: List[TransductionGraph] = []
                for tr, tier in zip(transducer.transducers, tg.tiers):
                    if composed_tiers:
                        composed_tiers.append(tier)
                    else:
                        if tr.in_lang == request.compose_from:
                            composed_tiers.append(tier)
                        else:
                            conv = Conversion(
                                in_lang=tr.in_lang,
                                out_lang=tr.out_lang,
                                substring_alignments=tier.substring_alignments(),
                            )
                            if request.indices:
                                conv.input_nodes = list(tier.input_string)
                                conv.output_nodes = list(tier.output_string)
                                conv.alignments = tier.alignments()
                            conversions.insert(0, conv)
                if composed_tiers:
                    composed_tg = CompositeTransductionGraph(composed_tiers)
                    conv = Conversion(
                        in_lang=request.compose_from,
                        out_lang=transducer.out_lang,
                        substring_alignments=composed_tg.substring_alignments(),
                    )
                    if request.indices:
                        conv.input_nodes = list(composed_tg.input_string)
                        conv.output_nodes = list(composed_tg.output_string)
                        conv.alignments = composed_tg.alignments()
                    conversions.insert(0, conv)
            else:
                for tr, tier in zip(transducer.transducers, tg.tiers):
                    conv = Conversion(
                        in_lang=tr.in_lang,
                        out_lang=tr.out_lang,
                        substring_alignments=tier.substring_alignments(),
                    )
                    if request.indices:
                        conv.input_nodes = list(tier.input_string)
                        conv.output_nodes = list(tier.output_string)
                        conv.alignments = tier.alignments()
                    conversions.insert(0, conv)
        segments.append(Segment(conversions=conversions))
    return segments


@api.get(
    "/outputs_for/{lang}",
    response_description="List of language codes into which {lang} can be converted",
)
def get_possible_output_conversions_for_a_writing_system(
    lang: LanguageNode = Path(description="Input language name"),
) -> List[str]:
    """Get the possible output languages for a given input language. These
    are all the phonetic or orthographic systems into which you can convert
    this input.
    """
    return sorted(g2p_langs.LANGS_NETWORK.descendants(lang.name))


@api.get(
    "/inputs_for/{lang}",
    response_description="List of language codes which can be converted into {lang}",
)
def get_writing_systems_that_can_be_converted_to_an_output(
    lang: LanguageNode = Path(description="Output language name"),
) -> List[str]:
    """Get the possible input languages for a given output language. These
    are all the phonetic or orthographic systems that you can convert
    into this output.
    """
    return sorted(g2p_langs.LANGS_NETWORK.ancestors(lang.name))


@api.get(
    "/path/{in_lang}/{out_lang}",
    response_description="Path from {in_lang} to {out_lang}",
)
def get_path_from_one_language_to_another(
    in_lang: LanguageNode = Path(description="Input language name"),
    out_lang: LanguageNode = Path(description="Output language name"),
) -> List[str]:
    """Get the sequence of intermediate forms used to convert from {in_lang} to {out_lang}."""
    try:
        return g2p_langs.LANGS_NETWORK.shortest_path(in_lang.name, out_lang.name)
    except ValueError:
        raise HTTPException(
            status_code=400, detail=f"No path from {in_lang} to {out_lang}"
        )



#!/usr/bin/env python

"""
Test all Mappings
"""

import io
from contextlib import redirect_stderr
from unittest import TestCase, main

from g2p.log import LOGGER
from g2p.mappings import Mapping
from g2p.mappings.create_ipa_mapping import (
    DISTANCE_METRICS,
    create_mapping,
    create_multi_mapping,
)
from g2p.transducer import Transducer


class MappingCreationTest(TestCase):
    def setUp(self):
        self.mappings = [
            {"in": "ɑ", "out": "AA"},
            {"in": "eː", "out": "EY"},
            {"in": "i", "out": "IY"},
            {"in": "u", "out": "UW"},
            {"in": "tʃ", "out": "CH"},
            {"in": "p", "out": "P"},
            {"in": "t", "out": "T"},
            {"in": "k", "out": "K"},
            {"in": "w", "out": "W"},
            {"in": "ɡ", "out": "G"},
            {"in": "ʒ", "out": "ZH"},
        ]
        self.mappings_xsampa = [
            {"in": "A", "out": "AA"},
            {"in": "e:", "out": "EY"},
            {"in": "i", "out": "IY"},
            {"in": "u", "out": "UW"},
            {"in": "tS", "out": "CH"},
            {"in": "p", "out": "P"},
            {"in": "t", "out": "T"},
            {"in": "k", "out": "K"},
            {"in": "w", "out": "W"},
            {"in": "g", "out": "G"},
            {"in": "Z", "out": "ZH"},
        ]
        self.target_mapping = Mapping(
            rules=self.mappings,
            in_lang="eng-ipa",
            out_lang="eng-arpabet",
            out_delimiter=" ",
        )
        self.target_mapping_xsampa = Mapping(
            rules=self.mappings_xsampa,
            in_lang="eng-xsampa",
            out_lang="eng-arpabet",
            out_delimiter=" ",
        )

    def test_unigram_mappings(self):
        src_mappings = [
            {"in": "ᐃ", "out": "i"},
            {"in": "ᐅ", "out": "u"},
            {"in": "ᐊ", "out": "a"},
        ]
        src_mapping = Mapping(rules=src_mappings, in_lang="crj", out_lang="crj-ipa")
        mapping = create_mapping(src_mapping, self.target_mapping, quiet=True)
        transducer = Transducer(mapping)
        self.assertEqual(transducer("a").output_string, "ɑ")
        self.assertEqual(transducer("i").output_string, "i")
        self.assertEqual(transducer("u").output_string, "u")

    def test_bigram_mappings(self):
        src_mappings = [
            {"in": "ᐱ", "out": "pi"},
            {"in": "ᑎ", "out": "ti"},
            {"in": "ᑭ", "out": "ki"},
        ]
        src_mapping = Mapping(rules=src_mappings, in_lang="crj", out_lang="crj-ipa")
        mapping = create_mapping(src_mapping, self.target_mapping, quiet=True)
        transducer = Transducer(mapping)
        self.assertEqual(transducer("pi").output_string, "pi")
        self.assertEqual(transducer("ti").output_string, "ti")
        self.assertEqual(transducer("ki").output_string, "ki")

    def test_trigram_mappings(self):
        src_mappings = [
            {"in": "ᒋ", "out": "t͡ʃi"},
            {"in": "ᒍ", "out": "t͡ʃu"},
            {"in": "ᒐ", "out": "t͡ʃa"},
        ]
        src_mapping = Mapping(rules=src_mappings, in_lang="crj", out_lang="crj-ipa")
        mapping = create_mapping(src_mapping, self.target_mapping, quiet=True)
        transducer = Transducer(mapping)
        self.assertEqual(transducer("t͡ʃi").output_string, "tʃi")
        self.assertEqual(transducer("t͡ʃu").output_string, "tʃu")
        self.assertEqual(transducer("t͡ʃa").output_string, "tʃɑ")

    def test_trigram_mappings_xsampa(self):
        src_mappings = [
            {"in": "ᒋ", "out": "tSi"},
            {"in": "ᒍ", "out": "tSu"},
            {"in": "ᒐ", "out": "tSa"},
        ]
        src_mapping = Mapping(rules=src_mappings, in_lang="crj", out_lang="crj-xsampa")
        mapping = create_mapping(src_mapping, self.target_mapping_xsampa, quiet=True)
        transducer = Transducer(mapping)
        self.assertEqual(transducer("tSi").output_string, "tSi")
        self.assertEqual(transducer("tSu").output_string, "tSu")
        self.assertEqual(transducer("tSa").output_string, "tSA")

    def test_long_mappings(self):
        src_mappings = [
            {"in": "ᐧᐯ", "out": "pʷeː"},
            {"in": "ᐧᑌ", "out": "tʷeː"},
            {"in": "ᐧᑫ", "out": "kʷeː"},
        ]
        src_mapping = Mapping(rules=src_mappings, in_lang="crj", out_lang="crj-ipa")
        mapping = create_mapping(src_mapping, self.target_mapping, quiet=True)
        transducer = Transducer(mapping)
        self.assertEqual(transducer("pʷeː").output_string, "pweː")
        self.assertEqual(transducer("tʷeː").output_string, "tweː")
        self.assertEqual(transducer("kʷeː").output_string, "kweː")

    def test_distance_errors(self):
        src_mappings = [{"in": "ᐃ", "out": "i"}]
        src_mapping = Mapping(rules=src_mappings, in_lang="crj", out_lang="crj-ipa")
        # Exercise looking up distances in the known list
        with self.assertRaises(ValueError):
            _ = create_mapping(
                src_mapping, self.target_mapping, distance="not_a_distance", quiet=True
            )
        with self.assertRaises(ValueError):
            _ = create_multi_mapping(
                [(src_mapping, "out")],
                [(self.target_mapping, "in")],
                distance="not_a_distance",
                quiet=True,
            )
        # White box testing: monkey-patch an invalid distance to validate the
        # second way we make sure distances are supported
        DISTANCE_METRICS.append("not_a_real_distance")
        with self.assertRaises(ValueError), self.assertLogs(LOGGER, level="ERROR"):
            _ = create_mapping(
                src_mapping,
                self.target_mapping,
                distance="not_a_real_distance",
                quiet=True,
            )
        with self.assertRaises(ValueError), self.assertLogs(LOGGER, level="ERROR"):
            _ = create_multi_mapping(
                [(src_mapping, "out")],
                [(self.target_mapping, "in")],
                distance="not_a_real_distance",
                quiet=True,
            )
        DISTANCE_METRICS.pop()

    def test_distances(self):
        # These mapppings are chosen to create different generated mappings
        # from the various distances.
        src_mappings = [
            {"in": "ᐧᐯ", "out": "pʷeː"},
            {"in": "ᒋ", "out": "t͡ʃi"},
            {"in": "ᕃ", "out": "ʁaj"},
        ]
        src_mapping = Mapping(rules=src_mappings, in_lang="crj", out_lang="crj-ipa")
        mapping = create_mapping(src_mapping, self.target_mapping, quiet=True)
        # print("mapping", mapping, list(mapping), "distance", "default")
        self.assertTrue(isinstance(mapping, Mapping))
        set_of_mappings = {tuple(rule.rule_output for rule in mapping.rules)}
        for distance in DISTANCE_METRICS:
            mapping = create_mapping(
                src_mapping, self.target_mapping, distance=distance, quiet=True
            )
            # print("mapping", mapping, list(mapping), "distance", distance)
            self.assertTrue(isinstance(mapping, Mapping))
            set_of_mappings.add(tuple(rule.rule_output for rule in mapping.rules))

            mapping = create_multi_mapping(
                [(src_mapping, "out")],
                [(self.target_mapping, "in")],
                distance=distance,
                quiet=True,
            )
            self.assertTrue(isinstance(mapping, Mapping))
            set_of_mappings.add(tuple(rule.rule_output for rule in mapping.rules))
        self.assertGreater(len(set_of_mappings), 3)

    def test_deletion_mapping(self):
        """Ensure that deletion rules do not lead to spurious warnings."""
        src_mappings = [
            {"in": "foo", "out": ""},
            {"in": "ᐃ", "out": "i"},
            {"in": "ᐅ", "out": "u"},
            {"in": "ᐊ", "out": "a"},
        ]
        src_mapping = Mapping(rules=src_mappings, in_lang="crj", out_lang="crj-ipa")
        log_output = io.StringIO()
        with redirect_stderr(log_output):
            mapping = create_mapping(src_mapping, self.target_mapping)
        self.assertFalse("WARNING" in log_output.getvalue())
        transducer = Transducer(mapping)
        self.assertEqual(transducer("a").output_string, "ɑ")
        self.assertEqual(transducer("i").output_string, "i")
        self.assertEqual(transducer("u").output_string, "u")


if __name__ == "__main__":
    main()



"""

Basic init file for g2p module

The main entry points for the g2p module are:
 - make_g2p() to create a mapper from and lang to another
 - make_tokenizer() to create a tokenizer for a given language
 - get_arpabet_langs() to get the list of languages with a path to eng-arpabet

Basic Usage:
    from g2p import make_g2p
    converter = make_g2p(in_lang, out_lang)
    transduction_graph = converter(input_text_in_in_alang)
    converted_text_in_out_lang = transduction_graph.output_string

    from g2p import make_tokenizer
    tokenizer = make_tokenizer(lang)
    for token in tokenizer.tokenize_text(input_text):
        if token.is_word:
            word = token.text
        else:
            interword_punctuation_and_spaces = token.text

    from g2p import get_arpabet_langs
    LANGS, LANG_NAMES = get_arpabet_langs()
"""

import sys
from typing import Dict, Optional, Tuple, Union

from g2p.exceptions import InvalidLanguageCode, NeuralDependencyError, NoPath
from g2p.shared_types import BaseTokenizer, BaseTransducer, Token

if sys.version_info < (3, 8):  # pragma: no cover
    sys.exit(
        "Python 3.8 or more recent is required by g2p.\n"
        f"You are using Python {sys.version}.\n"
        "Please use a newer version of Python."
    )

_g2p_cache: Dict[Tuple[str, str, bool, bool, int], BaseTransducer] = {}


def make_g2p(  # noqa: C901
    in_lang: str,
    out_lang: str,
    *,
    tokenize: bool = True,
    custom_tokenizer: Optional[BaseTokenizer] = None,
    neural: bool = False,
) -> BaseTransducer:
    """Make a g2p Transducer for mapping text from in_lang to out_lang via the
    shortest path between them.

    By default, the input is tokenized using the path of mappings from in_lang
    to out_lang, because transducers are not guaranteed to deal with whitespace,
    punctuation, and word boundaries properly.

    Args:
        in_lang (str): input language code
        out_lang (str): output language code
        tokenize (bool): whether tokenization should happen (default: True)
                         Warning: only set this to False if you've already
                         segmented your text into words without punctuation
        custom_tokenizer (Tokenizer): the tokenizer to use (default: a tokenizer
                                      built on the path from in_lang and out_lang)
        neural (bool): whether to use a neural model if it is available (default: False)

    Returns:
        Transducer from in_lang to out_lang, optionally with a tokenizer.

    Raises:
        InvalidLanguageCode: if in_lang or out_lang don't exist
        NoPath: if there is path between in_lang and out_lang
    """
    # Defer expensive imports
    from g2p.log import LOGGER
    from g2p.mappings import Mapping
    from g2p.mappings.langs import LANGS_NETWORK
    from g2p.mappings.utils import has_neural_support
    from g2p.transducer import CompositeTransducer, TokenizingTransducer, Transducer

    # Early checking for if neural is available:
    if neural and not has_neural_support():
        raise NeuralDependencyError()

    if (in_lang, out_lang, tokenize, neural, id(custom_tokenizer)) in _g2p_cache:
        return _g2p_cache[(in_lang, out_lang, tokenize, neural, id(custom_tokenizer))]

    # Check in_lang is a node in network
    if in_lang not in LANGS_NETWORK.nodes:
        LOGGER.error(f"No lang called '{in_lang}'. Please try again.")
        raise InvalidLanguageCode(in_lang)

    # Check out_lang is a node in network
    if out_lang not in LANGS_NETWORK.nodes:
        LOGGER.error(f"No lang called '{out_lang}'. Please try again.")
        raise InvalidLanguageCode(out_lang)

    if in_lang == out_lang:
        LOGGER.error(
            "Sorry, you can't transduce between the same language. "
            "Please select a different output language code."
        )
        raise NoPath(in_lang, out_lang)

    # Try to find the shortest path between the nodes
    try:
        path = LANGS_NETWORK.shortest_path(in_lang, out_lang)
    except ValueError:
        LOGGER.error(
            f"Sorry, we couldn't find a way to convert {in_lang} to {out_lang}. "
            "Please update your langs by running `g2p update` and try again."
        )
        raise NoPath(in_lang, out_lang)

    # Find all mappings needed
    mappings_needed = []
    for lang1, lang2 in zip(path[:-1], path[1:]):
        mapping = Mapping.find_mapping(in_lang=lang1, out_lang=lang2, neural=neural)
        LOGGER.debug(
            f"Adding mapping between {lang1} and {lang2} to composite transducer."
        )
        mappings_needed.append(mapping)

    # Either construct a Transducer or Composite Transducer
    transducer: Union[Transducer, CompositeTransducer, TokenizingTransducer]
    if len(mappings_needed) == 1:
        transducer = Transducer(mappings_needed[0])
    else:
        transducer = CompositeTransducer([Transducer(x) for x in mappings_needed])

    # If tokenization was requested, return a TokenizingTransducer
    if custom_tokenizer:
        transducer = TokenizingTransducer(transducer, custom_tokenizer)
    elif tokenize:
        tokenizer = make_tokenizer(in_lang=in_lang, tok_path=path)
        transducer = TokenizingTransducer(transducer, tokenizer)

    _g2p_cache[(in_lang, out_lang, tokenize, neural, id(custom_tokenizer))] = transducer
    return transducer


def tokenize_and_map(tokenizer: BaseTokenizer, transducer: BaseTransducer, input: str):
    result = ""
    for token in tokenizer.tokenize_text(input):
        if token.is_word:
            result += transducer(token.text).output_string
        else:
            result += token.text
    return result


_langs_cache = None
_lang_names_cache = None


def get_arpabet_langs():
    """Get the list of language codes and names supported by the g2p library
    for mapping to ARPABET.

    Example uses can be found in https://github.com/ReadAlongs/Studio and
    https://github.com/EveryVoiceTTS/EveryVoice

    Returns:
        LANGS (List[str]), LANG_NAMES (Dict[str,str]):
            LANGS is the sorted list of valid language codes supported
            LANG_NAMES maps each code to its full language name and is ordered by codes
    """
    # Defer expensive imports
    from g2p.mappings import LANGS
    from g2p.mappings.langs import LANGS_NETWORK

    global _langs_cache
    global _lang_names_cache

    if _langs_cache is not None and _lang_names_cache is not None:
        # Cache the results so we only calculate this information once.
        return _langs_cache, _lang_names_cache
    else:
        # langs_available in g2p lists langs inferred by the directory structure of
        # g2p/mappings/langs, but in ReadAlongs, we need all input languages to any mappings.
        # E.g., for Michif, we need to allow crg-dv and crg-tmd, but not crg, which is what
        # langs_available contains. So we define our own list of languages here.
        langs_available = []

        # this will be the set of all langs in g2p, which we need temporarily
        full_lang_names = {}

        for v in LANGS.values():
            for mapping in v.mappings:
                # add mapping to names hash table
                full_lang_names[mapping.in_lang] = mapping.language_name
                # add input id to all available langs list
                if mapping.in_lang not in langs_available:
                    langs_available.append(mapping.in_lang)

        # get the key from all networks in g2p module that have a path to 'eng-arpabet',
        # which is needed for the readalongs
        # Filter out <lang>-ipa: we only want "normal" input languages.
        # Filter out *-norm and crk-no-symbols, these are just intermediate representations.

        _langs_cache = [
            x
            for x in langs_available
            if not x.endswith("-ipa")
            and not x.endswith("-equiv")
            and not x.endswith("-no-symbols")
            and x not in ["und-ascii", "moh-festival"]
            and x in LANGS_NETWORK
            and LANGS_NETWORK.has_path(x, "eng-arpabet")
        ]

        # Sort LANGS so the -h messages list them alphabetically
        _langs_cache = sorted(_langs_cache)

        # Set up _lang_names_cache hash table for studio UI to properly name the dropdown options
        _lang_names_cache = {
            lang_code: full_lang_names[lang_code] for lang_code in _langs_cache
        }

        return _langs_cache, _lang_names_cache


def make_tokenizer(in_lang=None, out_lang=None, tok_path=None) -> BaseTokenizer:
    """Make the tokenizer for input in language in_lang

    Logic used when only in_lang is provided:
    - if in_lang -> in_lang-ipa, or in_lang -> X-ipa exists, tokenize using the input
      inventory of that mapping.
    - elif in_lang -> X -> Y-ipa exists, e.g., tce -> tce-equiv -> tce-ipa, tokenize
      using the input inventory of those two hops.
    - otherwise, just use the default tokenizer, which accepts as part of words all
      unicode letter, numbers and diacritics

    Logic used when in_lang and out_lang are provided:
    - if in_lang -> out_lang exists, tokenize using the input inventory of that mapping
    - otherwise use the default tokenizer

    Logic used when in_lang and tok_path are provided:
    - use the first one or two hops in path, stopping at the first -ipa node
    """
    from g2p.mappings.tokenizer import make_tokenizer as _make_tokenizer

    return _make_tokenizer(in_lang, out_lang, tok_path)


# Declare what's actually part of g2p's programmatic API.
# Please don't import anything else from g2p directly.
__all__ = [
    "BaseTokenizer",
    "BaseTransducer",
    "InvalidLanguageCode",
    "NoPath",
    "Token",
    "get_arpabet_langs",
    "make_g2p",
    "make_tokenizer",
    "tokenize_and_map",
]



#!/usr/bin/env python

"""
Test suite for the g2p-studio web app.

Requirements: python 3.8 and `test` dependencies

Before running this test suite, launch the g2p-studio server:
minimal dev mode:
    python run_studio.py
or robust server mode (*nix only, gunicorn does not work on Windows):
    gunicorn --worker-class uvicorn.workers.UvicornWorker -w 1 g2p.app:APP --bind 0.0.0.0:5000 --daemon
"""

from datetime import datetime
from random import sample
from unittest import IsolatedAsyncioTestCase, main

import socketio  # type: ignore
from playwright.async_api import async_playwright, expect  # type: ignore

from g2p.log import LOGGER
from g2p.tests.public.data import load_public_test_data


class StudioTest(IsolatedAsyncioTestCase):
    def __init__(self, *args):
        super().__init__(*args)
        self.port = 5000
        self.debug_convert = True
        self.timeout_delay = 500

    async def test_socket_connection(self):
        client = socketio.AsyncClient()
        await client.connect(
            f"http://127.0.0.1:{self.port}", socketio_path="/ws/socket.io"
        )
        await client.disconnect()

    async def test_sanity(self):
        async with async_playwright() as p:
            browser = await p.chromium.launch(channel="chrome", headless=True)
            page = await browser.new_page()
            await page.goto(f"http://127.0.0.1:{self.port}/docs")
            await page.wait_for_timeout(self.timeout_delay)
            await page.goto(f"http://127.0.0.1:{self.port}/static/swagger.json")
            await page.wait_for_timeout(self.timeout_delay)
            await page.goto(f"http://127.0.0.1:{self.port}")
            await page.wait_for_timeout(self.timeout_delay)
            input_el = page.locator("#input")
            output_el = page.locator("#output")
            await page.type("#input", "hello world")
            await page.wait_for_timeout(self.timeout_delay)
            input_text = await input_el.input_value()
            output_text = await output_el.input_value()
            self.assertEqual(input_text, output_text)
            self.assertEqual(input_text, "hello world")
            await input_el.fill("")
            await output_el.fill("")
            await page.type("#input", "hello world")
            await page.wait_for_timeout(self.timeout_delay)
            radio_el = page.locator("#animated-radio")
            await radio_el.click()
            await page.wait_for_timeout(self.timeout_delay)

    async def test_switch_langs(self):
        async with async_playwright() as p:
            browser = await p.chromium.launch(channel="chrome", headless=True)
            page = await browser.new_page()
            await page.goto(f"http://127.0.0.1:{self.port}")
            await page.wait_for_timeout(self.timeout_delay)
            await page.type("#input", "a")
            in_lang_selector = page.locator("#input-langselect")
            # Switch to a language
            await in_lang_selector.select_option(value="alq")
            await page.wait_for_timeout(self.timeout_delay)
            settings_title = await page.text_content("#link-0")
            self.assertEqual(settings_title, "Algonquin to IPA")
            # Switch output language
            out_lang_selector = page.locator("#output-langselect")
            await out_lang_selector.select_option("eng-arpabet")
            settings_title_3 = await page.text_content("#link-2")
            self.assertEqual(settings_title_3, "English IPA to Arpabet")
            # Switch back to custom
            await in_lang_selector.select_option(value="Custom")
            await page.wait_for_timeout(self.timeout_delay)
            settings_title = await page.text_content("#link-0")
            self.assertEqual(settings_title, "Custom")
            # FIXME: Test that the table works somewhere, somehow
            # Switch to in_lang = eng-arpabet, which means there is no possible outlang
            await in_lang_selector.select_option(value="eng-arpabet")
            await page.wait_for_timeout(self.timeout_delay)
            self.assertEqual(await page.locator("#link-0").count(), 0)

    async def test_langs(self):
        langs_to_test = load_public_test_data()
        # Doing the whole test set takes a long time, so let's use a 10% random sample,
        # knowing that all cases always get exercised in test_cli.py and test_langs.py.
        # 10% is enough to catch a sudden drift where the studio might stop being campatible.
        langs_to_test = [
            langs_to_test[i]
            for i in sorted(
                sample(range(len(langs_to_test)), k=len(langs_to_test) // 10)
            )
        ]
        # Make sure we test at least one lexicon-based example
        langs_to_test.append(["eng", "eng-arpabet", "hello", "HH AH L OW "])
        # Make sure case-insensitive lexicon-based mappings are applied case insensitive
        langs_to_test.append(["eng", "eng-arpabet", "HELLO", "HH AH L OW "])
        # Exercise a multi-word lexicon-based example now that we tokenize in studio
        langs_to_test.append(
            [
                "eng",
                "eng-arpabet",
                "Well, hello there!",
                "W EH L , HH AH L OW  DH EH R !",
            ]
        )
        # And with a Custom mapping with no rules in it, we just get passthrough
        langs_to_test.append(["custom", "custom", "foo bar", "foo bar"])

        # Known issue (#441): make sure "&"->"êkwa" is not exercised until we fix #441
        # This was causing CI failures we don't want to see until the bug is fixed.
        # TODO as part of the fix for #441, please delete this block of code
        def find_test_case(test_case, langs_to_test) -> int:
            """Return the index of test_case in langs_to_test if found, -1 otherwise."""
            for i, (in_lang, out_lang, input_text, output_text, *_) in enumerate(
                langs_to_test
            ):
                if [in_lang, out_lang, input_text, output_text] == test_case:
                    return i
            return -1

        known_failure = ["crk", "crk-no-symbols", "&", "êkwa"]
        known_failure_index = find_test_case(known_failure, langs_to_test)
        if known_failure_index != -1:
            LOGGER.warning(
                f"Ignoring record {known_failure} which is known to fail due to #441"
            )
            langs_to_test.pop(known_failure_index)
        # End of #441 work-around block of code

        error_count = 0

        # The current g2p-studio app leaks memory, so that if we try to run all the test
        # cases in one single browser instance, a case (not always the same) eventually
        # breaks. While that should get patched, for now, let's make unit testing
        # reliable by running tests by blocks we know the app can handle.
        block_size = 50

        max_action_delay = 5000  # in ms - max time we'll wait for an action to work
        polling_period = 20  # in ms - polling period for action effects

        for block in range((len(langs_to_test) - 1) // block_size + 1):
            LOGGER.info("Lauching async_playwright")
            async with async_playwright() as p:
                LOGGER.info(("L" if block == 0 else "Rel") + "aunching browser")
                browser = await p.chromium.launch(channel="chrome", headless=True)
                LOGGER.info("Loading page")
                page = await browser.new_page()
                await page.goto(
                    f"http://127.0.0.1:{self.port}", wait_until="networkidle"
                )

                # Define element locators
                input_el = page.locator("#input")
                output_el = page.locator("#output")
                in_lang_selector = page.locator("#input-langselect")
                out_lang_selector = page.locator("#output-langselect")
                table_status_selector = page.locator("#table-status")

                for i, test in enumerate(
                    langs_to_test[block_size * block : block_size * (block + 1)]
                ):
                    test_in_lang = test[0].strip()
                    test_out_lang = test[1].strip()
                    test_input_text = test[2].strip()
                    test_expected_output = test[3].strip()
                    LOGGER.info(
                        f"{i + block * block_size} {datetime.now()} "
                        f"{test_in_lang}->{test_out_lang} {test_input_text} -> {test_expected_output}"
                    )
                    timed_out_count = 0
                    for attempt in range(1, 4):
                        if attempt > 1:
                            LOGGER.info(f"Attempt #{attempt}")
                            await page.wait_for_timeout(self.timeout_delay)
                        # Clear input and output
                        await input_el.fill("")
                        await output_el.fill("")
                        output_text = ""

                        # Select the input language
                        # descendants_promise = page.wait
                        async with page.expect_request(
                            "**/descendants/" + test_in_lang
                        ) as req_info:
                            await in_lang_selector.select_option(value=test_in_lang)
                        await req_info.value
                        # wait up to max_action_delay ms for input lang to be set
                        # and for mappings to be populated give it maximum 20s to show up and fail
                        await expect(table_status_selector).to_have_attribute(
                            "state", "loaded", timeout=20000
                        )

                        # Select the output language
                        await out_lang_selector.select_option(value=test_out_lang)
                        # wait up to max_action_delay ms for output lang to be set
                        # and for mappings to be populated
                        await page.wait_for_timeout(polling_period)
                        await expect(table_status_selector).to_have_attribute(
                            "state", "loaded", timeout=20000
                        )

                        # Type fill input, then trigger rendering with keyup event
                        # optimization: make sure there is only 1 keyup event
                        await input_el.fill(test_input_text + " ")
                        await input_el.press("Backspace")

                        loop_time = 0
                        # wait up to max_action_delay ms for output to be populated
                        while loop_time <= max_action_delay:
                            output_text = await output_el.input_value()
                            if output_text.strip() == test_expected_output:
                                break
                            await page.wait_for_timeout(polling_period)
                            loop_time += polling_period
                        else:
                            LOGGER.warning(
                                f"Reached timeout setting input text for {test}"
                            )
                            timed_out_count += 1
                            continue

                        # We're done trying once an attempt succeeds
                        if output_text.strip() == test_expected_output:
                            break

                    if (
                        timed_out_count == 3
                        and output_text.strip() != test_expected_output
                    ):  # pragma: no cover
                        output_text = output_text + " (TIMED OUT)"
                    # Check that output is correct after the first succesful attempt or
                    # after all the attempts have failed.
                    if not self.debug_convert:
                        self.assertEqual(output_text.strip(), test_expected_output)
                        LOGGER.info(
                            f"Successfully converted {test_input_text} from {test_in_lang} to {test_out_lang}"
                        )
                    elif output_text.strip() != test_expected_output:
                        LOGGER.warning(
                            f"test_langs.py: mapping error: {test_input_text} from {test_in_lang} "
                            f"to {test_out_lang} should be {test_expected_output}, got {output_text}"
                        )
                        if error_count == 0:
                            first_failed_test = [test_expected_output, output_text]
                        error_count += 1
                    else:
                        LOGGER.info(
                            f"Successfully converted {test_input_text} from {test_in_lang} to {test_out_lang}"
                        )

                # Let the user know we're closing the browser (by exiting the p context
                # manager scope) to explain why there's a delay here.
                LOGGER.info("Closing browser")

        if self.debug_convert and error_count > 0:
            self.assertEqual(
                first_failed_test[0],
                first_failed_test[1],
                f"{error_count} lang mapping test case(s) failed, "
                "look for warnings in the logs above for details.",
            )


if __name__ == "__main__":
    main()
